#include "patch.h"

/*
	DALLOC.C -- Malloc wrappers

	Based on sources from Lee Killough. This module defines dynamic
	memory calls that track the allocations and mark the blocks with
	a signature. When NOWRAPPER is defined these routines compile into
	calls to the normal malloc, free, and realloc calls. This provides

	xmalloc:  Traps failed allocations, records owner, size, address
	xfree:    Traps frees of unallocated memory, deletes owner, size, address
	xrealloc: Traps realloc of non NULL unallocated memory, chg owner size addr

	the above routines are never called explicitly, they are used SOLELY by
	the macros free, malloc, and realloc.

	xatexit:  Installs exit handler that prints out remaining un-freed blocks.
	xlibfree: Provides access to original library free for library allocations

	#include dalloc.h to define the memory allocation wrappers in all source
	modules that use dynamic memory, and include dalloc.c in the compile.
	You can then turn the wrappers on and off by defining NOWRAPPER or not.
*/

#define SIG 0xF2E765A9
#define SIG_SIZE 16

// define the structure for information about a dynamic allocation

static struct plist
{
	struct plist *next;		// linked list of allocations, next list item
	void *ptr;						// address of allocated block
	size_t size;					// requested size of allocation	
	int lno;							// source line number of allocation request
	char fnam[16];				// source file name of allocation request
} *alloced = NULL;

// do not use wrappers in this module, this module DEFINES the wrappers

#undef free
#undef malloc
#undef realloc

// local storage for wrappers 
// flag for leak report disable on error, current and peak allocations

static int no_memleak = 0;  // flag to turn off leak display when memory error
static int tot_alloc = 0;		// total bytes allocated unwrapped at given time
static int pk_alloc = 0;		// records largest total unwrapped allocation
static int tot_wralloc = 0;	// total bytes allocated wrapped at given time
static int pk_wralloc = 0;	// total bytes allocated wrapped at given time

//
// xmalloc -- malloc wrapper
//
// allocate a block of the given size, plus a SIG_SIZE signature
// allocate an allocation info descriptor and append it to allocation list
	
void *xmalloc(char *fnam,int lno,size_t n)
{
	char *p;
	
	// unwrapped malloc

#ifdef NOWRAPPER
	p = malloc(n);				// if not wrapped just allocate normally
	assert(p);					// and abort if fails
	return p;					// return pointer to alloced memory or NULL
#endif

	// wrapped malloc

	if (!n)						// report attempt to malloc 0 bytes
	{
     fprintf(stderr,"Malloc of zero length in %s at line %d. Use free,\n",fnam,lno);
		 no_memleak=1;
     exit(2);
	}
	p = malloc(n+SIG_SIZE);		// malloc for requested block plus signature

	// if the malloc succeeeded record the allocation in the list of such

	if (p)
	{
		struct plist *lp = malloc(sizeof(*lp));		// new list node
		if (!lp)									// error if can't alloc
		{
	     fprintf(stderr,"Malloc error in %s at line %d. Could not allocate memory for trace, size: %ld\n",fnam,lno,n);
			 no_memleak=1;
	     exit(2);
		}
		
		tot_alloc += n;								// record allocation
		tot_wralloc += n+SIG_SIZE+sizeof(*lp);		// size changes
		if (tot_alloc > pk_alloc)
			pk_alloc = tot_alloc;
		if (tot_wralloc > pk_wralloc)
			pk_wralloc = tot_wralloc;
		
		lp->next = alloced;							// append new at start
		lp->size = n;
		lp->ptr = p;
		strncpy(lp->fnam,fnam,15);
		lp->fnam[15]='\0';
		lp->lno = lno;
		alloced = lp;
		
		*(unsigned *) p = SIG;
		p += SIG_SIZE;
	}
	else
	{
	     fprintf(stderr,"Malloc failed in %s at line %d. Could not allocate memory, size: %ld\n",fnam,lno,n);
		 no_memleak=1;
	     exit(2);
	}
	return p;
}

// free wrapper
		
void xfree(char *fnam,int lno,void *p)
{
#ifdef NOWRAPPER
	return free(p);
#endif

	if (p)
	{
		struct plist *t, **lp = &alloced;

		p = (char *) p - SIG_SIZE;

		while (*lp && (*lp)->ptr != p)
			lp = &(*lp)->next;

		if (!*lp || *(unsigned *) p != SIG )
		{
			if (*lp)
				fprintf(stderr,"Free error in %s at line %d -- Alloced in %s at line %d @0x%lX\n",fnam,lno,(*lp)->fnam,(*lp)->lno,(unsigned long) p);
			else
				fprintf(stderr,"Free error in %s at line %d -- Block not allocated @0x%lX\n",fnam,lno,(unsigned long) p);
			no_memleak=1;
			exit(2);
		}
		tot_alloc -= (*lp)->size;
		tot_wralloc -= (*lp)->size - SIG_SIZE - sizeof(*lp);
		t = (*lp)->next;
		free(*lp);
		*lp = t;
		*(unsigned *) p = 0;
		free(p);
	}
}

// realloc wrapper - xmalloc, copy and xfree old

void *xrealloc(char *fnam,int lno,void *p,size_t n)
{
	char *q;
	struct plist *lp;

#ifdef NOWRAPPER
	q = realloc(p,n);
	if (n) assert(q);
	return q;
#endif

	q = xmalloc(fnam,lno,n);
	if (!p) return q;

	if (!n)
	{
		xfree(fnam,lno,p);
		return NULL;
	}

	for (lp=alloced; lp && lp->ptr != ((char *)p)-SIG_SIZE; lp=lp->next)
		;
	if (lp)
	{
		memcpy(q,p,n<lp->size? n : lp->size);
		xfree(fnam,lno,p);
		return q;
	}

	fprintf(stderr,"Realloc error in %s at line %d -- Block not allocated\n",fnam,lno);
	no_memleak=1;
	exit(2);
}

// exit handler that displays memory leaks
// if an error occurs no_memleak is turned on, so leaks don't obscure error

void xatexit(void)
{
#ifdef NOWRAPPER
	return;
#endif
	printf("Peak allocation: %8d  Peak wrapped allocation: %8d\n",pk_alloc,pk_wralloc);
	if (!no_memleak && alloced)
	{
		struct plist *lp;
		fprintf(stderr,"Memory leaks:\n");
		for (lp=alloced; lp; lp=lp->next)
		{
			unsigned char *p = lp->ptr;
			size_t i,n;

			fprintf(stderr,"0x%lX,from: %12s line: %5d size: %8X\n",
				(unsigned long)p,lp->fnam,lp->lno,(unsigned)lp->size);
			fprintf(stderr,"data: ");
			n = lp->size < 16 ? lp->size : 16;
			for (i=0; i<n; i++)
				fprintf(stderr," %02X", (unsigned) p[SIG_SIZE+i]);
			fprintf(stderr,"\n");
		}
	}
}

// report whether any memory is still allocated

int xalloced()
{
	return alloced? 1 : 0;
}

// use the library free routine for library allocations

void xlibfree(void *p)
{
#undef free
	free(p);
}
