#include "patch.h"


typedef struct boom_201_header
{
	unsigned char version;
	char signature[6];

	unsigned char compatibility;
	unsigned char gameskill;
	unsigned char gameepisode;
	unsigned char gamemap;
	unsigned char deathmatch;
	unsigned char consoleplayer;

	// option base
	unsigned char monsters_remember;
	unsigned char variable_friction;
	unsigned char weapon_recoil;
	unsigned char allow_pushers;
	unsigned char zero;
	unsigned char player_bobbing;
	unsigned char respawnparm;
	unsigned char fastparm;
	unsigned char nomonsters;
	unsigned char demo_insurance;
	unsigned char random_seed3;
	unsigned char random_seed2;
	unsigned char random_seed1;
	unsigned char random_seed0;

	unsigned char padding[64-14];

	// 128 from option base in v2.01, 256 from option base in v2.00
	unsigned char playeringame[32];

} BOOM201Header;

typedef struct boom_200_header
{
	unsigned char version;
	char signature[6];

	unsigned char compatibility;
	unsigned char gameskill;
	unsigned char gameepisode;
	unsigned char gamemap;
	unsigned char deathmatch;
	unsigned char consoleplayer;

	// option base
	unsigned char monsters_remember;
	unsigned char variable_friction;
	unsigned char weapon_recoil;
	unsigned char allow_pushers;
	unsigned char zero;
	unsigned char player_bobbing;
	unsigned char respawnparm;
	unsigned char fastparm;
	unsigned char nomonsters;
	unsigned char demo_insurance;
	unsigned char random_seed3;
	unsigned char random_seed2;
	unsigned char random_seed1;
	unsigned char random_seed0;

	unsigned char padding[128-14];

	unsigned char playeringame[32];

} BOOM200Header;

typedef struct doom_190_header
{
	unsigned char version;
	unsigned char gameskill;
	unsigned char gameepisode;
	unsigned char gamemap;
	unsigned char deathmatch;
	unsigned char respawnparm;
	unsigned char fastparm;
	unsigned char nomonsters;
	unsigned char consoleplayer;
	unsigned char playeringame[4];

} DOOM190Header;

typedef struct doom_120_header
{
	unsigned char gameskill;
	unsigned char gameepisode;
	unsigned char gamemap;
	unsigned char playeringame[4];

} DOOM120Header;

typedef enum
{
	V120,
	V190,
	V200,
	V201
} DemoTypes;

int bitcnt16[16]={0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4};
int ConvertVersion=0;
int ConvertLevel=0;
int ShowHeader=1;
int ShowMoves=0;
int FromWad=0;
int TruncDemo=0;
int StripPause=0;
char *demobuff=NULL;
char *long_skills[]=
{
	"Too Young to Die",
	"Not Too Rough",
	"Hurt Me Plenty",
	"Ultraviolent",
	"Nightmare"
};

//-------------------------------------------------------------
// Demo class member function definitions - Demo mode script
//
// A demo is a record of a game played, and consists of a record
// of game conditions, and a list of moves and actions made by the
// player during the demo recording. The DEMO1, DEMO2, and DEMO3
// resources may contain these recordings.
//-------------------------------------------------------------

// Return an integer indicating demo version
// 104, 105, or 106 indicates long header version, 0-4 indicates DOOM v1.2 demo
// 200, 201, or 202 indicates a BOOM format demo

int GetDemoVersion(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	return cp[0];
}

// Return an enumerated type correspoding to the demo format

DemoTypes GetDemoType(char *data)
{
	int dv = GetDemoVersion(data);

	if (dv<=4) return V120;
	if (dv<=109) return V190;
	if (dv<=200) return V200;
	return V201;
}

// Check the signature of the demo, return if valid, else null

const char *GetDemoSignature(char *data)
{
	UBYTE *cp;
	static char sig[8];

	cp = (UBYTE *)data;
	switch(GetDemoType(data))
	{
	  case V120: case V190:
			return "None";
		case V200: case V201:
			if (cp[1]!=0x1d || cp[6]!=0xe6)
				return "Error: bounding bytes not 0x1d/0xe6";
			if (strncmp(cp+2,"Boom",4))
			  return "Error: Signature corrupt";
			strncpy(sig,cp+2,4);
			sig[4]=0;
			return sig;
	}
	return NULL;
}

// Return an integer indicating skill level played
// 0=Too Young to Die,... 3=UV, 4=Nightmare

int GetSkillLevel(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[0];
		case V190:
			return cp[1];
		case V200: case V201:
			return cp[7];
	}
	return -1;
}

// Return an integer indicating episode played
// DOOM I returns 1, 2, or 3
// DOOM II always returns 1

int GetEpisode(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[1];
		case V190:
			return cp[2];
		case V200: case V201:
			return cp[8];
	}
	return -1;
}

// Return an integer indicating level played
// DOOM I returns 1-9
// DOOM II returns 1-32

int GetMission(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[2];
		case V190:
			return cp[3];
		case V200: case V201:
			return cp[9];
	}
	return -1;
}

// Return an integer indicating mode played in
// 0 = single/coop 1 = deathmatch 2 = altdeath
// v1.2 did not record mode, -1 is returned then

int GetMode(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return -1;
		case V190:
			return cp[4];
		case V200: case V201:
			return cp[10];
	}
	return -1;
}

// Return an integer indicating if Respawn was on
// 0 = Respawn off 1 = Respawn on
// v1.2 did not record Respawn, -1 is returned then

int GetRespawn(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return -1;
		case V190:
			return cp[5];
		case V200: case V201:
			return cp[19];
	}
	return -1;
}

// Return an integer indicating if Fast was on
// 0 = Fast off 1 = Fast on
// v1.2 did not record Fast, -1 is returned then

int GetFast(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return -1;
		case V190:
			return cp[6];
		case V200: case V201:
			return cp[20];
	}
	return -1;
}

// Return an integer indicating if NoMonsters was on
// 0 = NoMonsters off 1 = NoMonsters on
// v1.2 did not record NoMonsters, -1 is returned then

int GetNoMonsters(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return -1;
		case V190:
			return cp[7];
		case V200: case V201:
			return cp[21];
	}
	return -1;
}

int GetCompatibility(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			return cp[7];
		default:
			return -1;
	}
	return -1;
}

int GetMonstersRemember(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			return cp[13];
		default:
			return -1;
	}
	return -1;
}

int GetVariableFriction(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			return cp[14];
		default:
			return -1;
	}
	return -1;
}

int GetWeaponRecoil(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			return cp[15];
		default:
			return -1;
	}
	return -1;
}

int GetAllowPushers(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			return cp[16];
		default:
			return -1;
	}
	return -1;
}

int GetPlayerBobbing(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			return cp[18];
		default:
			return -1;
	}
	return -1;
}

int GetDemoInsurance(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			return cp[22];
		default:
			return -1;
	}
	return -1;
}

unsigned long GetRandomSeed(char *data)
{
	UBYTE *cp;
	unsigned long seed=0;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			seed = cp[23];
			seed <<=8;
			seed |= cp[24];
			seed <<=8;
			seed |= cp[25];
			seed <<=8;
			seed |= cp[26];
			break;
		default:
			seed = 0;
	}
	return seed;
}



// Return an integer indicating demo viewpoint
// 0 = Player 1,... 3 = Player 4
// v1.2 did not record viewpoint, -1 is returned then

int GetViewPoint(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return -1;
		case V190:
			return cp[8];
		case V200: case V201:
			return cp[12];
	}
	return -1;
}

// Return an integer indicating if Player 1 was present
// 0 = Player 1 not present, 1 = Player 1 was present

int HasPlayer1(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[3];
		case V190:
			return cp[9];
		case V200: 
			return cp[128+13];
		case V201:
			return cp[64+13];
	}
	return -1;
}

// Return an integer indicating if Player 2 was present
// 0 = Player 2 not present, 1 = Player 2 was present

int HasPlayer2(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[4];
		case V190:
			return cp[10];
		case V200: 
			return cp[128+13+1];
		case V201:
			return cp[64+13+1];
	}
	return -1;
}

// Return an integer indicating if Player 3 was present
// 0 = Player 3 not present, 1 = Player 3 was present

int HasPlayer3(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[5];
		case V190:
			return cp[11];
		case V200: 
			return cp[128+13+2];
		case V201:
			return cp[64+13+2];
	}
	return -1;
}

// Return an integer indicating if Player 4 was present
// 0 = Player 4 not present, 1 = Player 4 was present

int HasPlayer4(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[6];
		case V190:
			return cp[12];
		case V200: 
			return cp[128+13+3];
		case V201:
			return cp[64+13+3];
	}
	return -1;
}

// Return an integer indicating if Player N was present N<32
// 0 = Player N not present, 1 = Player N was present

int HasPlayerN(int n,char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return cp[3+n];
		case V190:
			return cp[9+n];
		case V200: 
			return cp[128+13+n];
		case V201:
			return cp[64+13+n];
	}
	return -1;
}

// Return the offset to the motion data

int MoveOffset(char *data)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V120:
			return 7;
		case V190:
			return 13;
		case V200: 
			return 128+13+32;
		case V201:
			return 64+13+32;
	}
	return -1;
}

// Return the number of gametics moves are stored for in the resource

int NumberTics(char *data,int len)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	return (len - 1 - MoveOffset(data))/4; // reslen - endbyte - header len
}

// Return a pointer to the first gametic of the demo

SBYTE *FirstTic(char *data)
{
	SBYTE *cp;

	cp = (SBYTE *)data;

	return cp+MoveOffset(data);
}

// Return forward/backward, strafe, turn, and shoot/push/switch info

void GetMove(char *data,int ticno,SBYTE *fbmove,SBYTE *slrmove,SBYTE *tlrmove,UBYTE *action)
{
	UBYTE *cp;
	SBYTE *mp;

	cp = (UBYTE *)data;
	cp += 4*ticno + MoveOffset(data);  // cp-> ticno-th move
	mp = (SBYTE *)cp; 									// mp-> ticno-th move

	*fbmove = mp[0];
	*slrmove = mp[1];
	*tlrmove = mp[2];
	*action = cp[3];
}

// Return forward/backward motion

SBYTE GetFBMotion(char *data,int ticno)
{
	SBYTE *mp;

	mp = (SBYTE *)data;
	mp += 4*ticno + MoveOffset(data);				// mp-> ticno-th move

	return mp[0];
}

// Return strafe motion

SBYTE GetSLRMotion(char *data,int ticno)
{
	SBYTE *mp;

	mp = (SBYTE *)data;
	mp += 4*ticno + MoveOffset(data);				// mp-> ticno-th move

	return mp[1];
}

// Return turn motion

SBYTE GetTLRMotion(char *data,int ticno)
{
	SBYTE *mp;

	mp = (SBYTE *)data;
	mp += 4*ticno + MoveOffset(data);	// mp-> ticno-th move

	return mp[2];
}

// Return shoot/push/switch info

UBYTE GetActions(char *data,int ticno)
{
	UBYTE *mp;

	mp = (UBYTE *)data;
	mp += 4*ticno + MoveOffset(data);	// mp-> ticno-th move

	return mp[3];
}

// Set Demo Version 104 or higher

void SetDemoVersion(char *data,int dv)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	if (GetDemoType(data)!=V120)
		cp[0]=dv;
}

// Set Skill for Demo Version 104 or higher
// 0=Too Young to Die,... 3=UV, 4=Nightmare

void SetSkillLevel(char *data,int sk)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[1]=sk;
}

// Set Episode for Demo Version 104 or higher
// DOOM I uses 1, 2, or 3
// DOOM II always sets 1

void SetEpisode(char *data,int ep)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[2]=ep;
}

// Set Mission for Demo Version 104 or higher
// DOOM I uses 1-9
// DOOM II sets 1-32

void SetMission(char *data,int ms)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[3]=ms;
}

// Set Mode for Demo Version 104 or higher
// 0 = single/coop 1 = deathmatch 2 = altdeath

void SetMode(char *data,int md)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[4]=md;
}

// Set Respawn for Demo Version 104 or higher
// 0 = Respawn off 1 = Respawn on

void SetRespawn(char *data,int rs)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[5]=rs;
}

// Set Fast for Demo Version 104 or higher
// 0 = Fast off 1 = Fast on

void SetFast(char *data,int fa)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[6]=fa;
}

// Set NoMonsters for Demo Version 104 or higher
// 0 = NoMonsters off 1 = NoMonsters on

void SetNoMonsters(char *data,int nm)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[7]=nm;
}

// Set Viewpoint for Demo Version 104 or higher
// 0 = Player 1,... 3 = Player 4

void SetViewPoint(char *data,int vp)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[8]=vp;
}

// Set Player 1 presence for Demo Version 104 or higher
// 0 = Player 1 not present, 1 = Player 1 is present

void SetHasPlayer1(char *data,int p1)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[9]=p1;
}

// Set Player 2 presence for Demo Version 104 or higher
// 0 = Player 2 not present, 1 = Player 2 is present

void SetHasPlayer2(char *data,int p2)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[10]=p2;
}

// Set Player 3 presence for Demo Version 104 or higher
// 0 = Player 3 not present, 1 = Player 3 is present

void SetHasPlayer3(char *data,int p3)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[11]=p3;
}

// Set Player 4 presence for Demo Version 104 or higher
// 0 = Player 4 not present, 1 = Player 4 is present

void SetHasPlayer4(char *data,int p4)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	cp[12]=p4;
}

// Change the length of the resource by adding or removing gametic
// entries from the end.

void SetNumberTics(char **data,int ntics,int *len)
{
	UBYTE *cp,v;
	int oldntics;

	cp = (UBYTE *)(*data);
	v = cp[0];

	oldntics = ((*len) - 1 - (MoveOffset(*data)))/4;
	if (oldntics!=ntics)
	{
		(*data) = (char *)realloc((*data),((*len)=(1+4*ntics+(MoveOffset(*data)))));
		(*data)[MoveOffset(*data)+ntics*4]=128;
	}

}
void SetCompatibility(char *data,int c)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[7] = c;
		default:
			break;
	}
}

void SetMonstersRemember(char *data,int c)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[13] = c;
		default:
			break;
	}
}

void SetVariableFriction(char *data,int c)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[14] = c;
		default:
			break;
	}
}

void SetWeaponRecoil(char *data,int c)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[15] = c;
		default:
			break;
	}
}

void SetAllowPushers(char *data,int c)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[16] = c;
		default:
			break;
	}
}

void SetPlayerBobbing(char *data,int c)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[18] = c;
		default:
			break;
	}
}

void SetDemoInsurance(char *data,int c)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[22] = c;
		default:
			break;
	}
}

void SetRandomSeed(char *data,unsigned long s)
{
	UBYTE *cp;

	cp = (UBYTE *)data;

	switch(GetDemoType(data))
	{
		case V200: case V201:
			cp[26] = s&255;
			cp[25] = (s>>8)&255;
			cp[24] = (s>>16)&255;
			cp[23] = (s>>24)&255;
		default:
			break;
	}
}

void RemovePauses(char **data,int *len)
{
	UBYTE *cp,v,*t0,*t,*p0;
	int ntics,ln=*len,hl;

	cp = (UBYTE *)(*data);
	v = *cp;
	hl = MoveOffset(*data);
	t0 = cp+hl;	// t0 points to first byte of first gametic

	ntics = (ln-1-hl)/4;

	// v               0    1    2    3    4    5    6    
	// hhhhhhhhhhhhh   xxxx xxxy xxxy xxxy xxxx xxxy xxxx z-
	// cp        cp+hl=t0	t			   t			   cp+ln


	t=t0;
	do				  		
	{
		for (;(t-cp<ln-1) && t[3]!=0x81;t+=4)
			;

		p0 = t;				// p0 pointer to first pause
//		dprintf(logfile,"Starting pause at %d\n",t-cp);
		if (t-cp<ln-1)
		{
			for (t+=4;(t-cp<ln-1) && t[3]!=0x81;t+=4)
				;			// pointer to second pause or end byte
			if (t[3]==0x81) t+=4;
//			dprintf(logfile,"Ending pause at %d\n",t-cp);
			if (t-cp<ln-1)
				memmove(p0,t,ln-(t-cp));
		}
		else break;

		ln -= t-p0;
		t = p0;
	}
	while (t-cp<ln-1);

	(*data) = (char *)realloc((*data),((*len)=ln));
}

// Change the length of the resource by adding gametic
// entries from the end.

void AddTics(char **data,int ntics,int *len)
{
	UBYTE *cp,v;

	cp = (UBYTE *)(*data);
	v = cp[0];

	if (ntics>0)
		(*data) = (char *)realloc((*data),(*len)=((*len)+ntics*4));
	if ((*data)==NULL) (*len)=0;
}

// Change the length of the resource by deleting gametic
// entries from startic, and for ntics moves.

void RemoveTics(char **data,int ntics,int starttic,int *len)
{
	UBYTE *cp,v;
	int oldntics,tic0;
	int i,j;

	cp = (UBYTE *)(*data);
	v = cp[0];
	tic0 = MoveOffset(*data);
	oldntics = ((*len) - 1 - tic0)/4;

	if (ntics>0)
	{
		for (i=starttic;i<oldntics-ntics;i++)
			for (j=0;j<4;j++)
				cp[starttic+4*i+j] = cp[starttic+4*i+j+ntics*4];

		(*data) = (char *)realloc((*data),(*len)=((*len)-4*ntics));
	}
}

// Set all features of one gametic move

void SetMove(char *data,int ticno,SBYTE fbmove,SBYTE slrmove,SBYTE tlrmove,UBYTE action)
{
	UBYTE *cp;
	SBYTE *mp;

	cp = (UBYTE *)data;
	cp += 4*ticno + MoveOffset(data);	// cp-> ticno-th move
	mp = (SBYTE *)cp; 								// mp-> ticno-th move

	mp[0] = fbmove;
	mp[1] = slrmove;
	mp[2] = tlrmove;
	cp[3] = action;
}

// Set forward/backward motion of one gametic move

void SetFBMotion(char *data,int ticno,SBYTE fbmove)
{
	SBYTE *mp;

	mp = (SBYTE *)data;
	mp += 4*ticno + (mp[0]<=4? 7 : 13);	// mp-> ticno-th move

	mp[0] = fbmove;
}

// Set strafe left/right of one gametic move

void SetSLRMotion(char *data,int ticno,SBYTE slrmove)
{
	SBYTE *mp;

	mp = (SBYTE *)data;
	mp += 4*ticno + MoveOffset(data);	// mp-> ticno-th move

	mp[1] = slrmove;
}

// Set turn left/right of one gametic move

void SetTLRMotion(char *data,int ticno,SBYTE tlrmove)
{
	SBYTE *mp;

	mp = (SBYTE *)data;
	mp += 4*ticno + MoveOffset(data);	// mp-> ticno-th move

	mp[2] = tlrmove;
}

// Set actions for one gametic move

void SetActions(char *data,int ticno,UBYTE action)
{
	UBYTE *mp;

	mp = (UBYTE *)data;
	mp += 4*ticno + MoveOffset(data);	// mp-> ticno-th move

	mp[3] = action;
}

char demonumber(char *vername)
{
	if (strstr(vername,"1.6")) return 106;	
	if (strstr(vername,"1.7")) return 107;	
	if (strstr(vername,"1.8")) return 108;	
	if (strstr(vername,"1.9")) return 109;
	if (strstr(vername,"2.0")) return 200;
	if (strstr(vername,"2.00")) return 200;
	if (strstr(vername,"2.01")) return 201;
	if (strstr(vername,"2.1")) return 201;
	if (strstr(vername,"2.02")) return 202;
	if (strstr(vername,"2.2")) return 202;
	return 109;
}

char *Duration(int dur)
{
	int hr,min,sec;
	static char out[32]="";

	hr = (dur/60)/60;
	min = (dur/60)%60;
	sec = dur%60;
	sprintf(out,"%1d:%02d:%02ds",hr,min,sec);
	return out;
}

void DumpHeader(char *db,int len,char *resname)
{
	int pmsk;
	int dur;
	DemoTypes t;

	t = GetDemoType(db);
	dprintf(logfile,"\nDumping %s header\n\n",resname);
	dprintf(logfile,"Demo Version Number %0d.%0d\n",GetDemoVersion(db)/100,GetDemoVersion(db)%100);
	if (t==V200 || t == V201)
		dprintf(logfile,"Signature: %s\n",GetDemoSignature(db));
	dprintf(logfile,"Skill Level: %s\n",long_skills[GetSkillLevel(db)]);
	dprintf(logfile,"Episode Played: %d\n",GetEpisode(db));
	dprintf(logfile,"Mission Played: %d\n",GetMission(db));
	if (t!=V120)
	{
		dprintf(logfile,"Play Mode: %s\n",GetMode(db)==0? "single/coop" : GetMode(db)==1? "deathmatch" : "altdeath");
		dprintf(logfile,"Respawn Mode: %s\n",GetRespawn(db)? "Yes" : "No");
		dprintf(logfile,"Fast Mode: %s\n",GetFast(db)? "Yes" : "No");
		dprintf(logfile,"No Monsters Mode: %s\n",GetNoMonsters(db)? "Yes" : "No");
		dprintf(logfile,"Player Viewpoint: Player %d\n",GetViewPoint(db)+1);
	}
	if (t==V200 || t==V201)
	{
		dprintf(logfile,"Compatibility Mode: %s\n",GetCompatibility(db)? "Yes" : "No");
		dprintf(logfile,"Monsters Remember: %s\n",GetMonstersRemember(db)? "Yes" : "No");
		dprintf(logfile,"Variable Friction: %s\n",GetVariableFriction(db)? "Yes" : "No");
		dprintf(logfile,"Weapon Recoil: %s\n",GetWeaponRecoil(db)? "Yes" : "No");
		dprintf(logfile,"Allow Pushers: %s\n",GetAllowPushers(db)? "Yes" : "No");
		dprintf(logfile,"Player Bobbing: %s\n",GetPlayerBobbing(db)? "Yes" : "No");
		dprintf(logfile,"Demo Insurance: %s\n",GetDemoInsurance(db)? GetDemoInsurance(db)-1? "When Recording" : "Yes" : "No");
		dprintf(logfile,"Seed: %ld\n",GetRandomSeed(db));
	}
	dprintf(logfile,"Player 1 Present: %s\n",HasPlayer1(db)? "Yes" : "No");
	dprintf(logfile,"Player 2 Present: %s\n",HasPlayer2(db)? "Yes" : "No");
	dprintf(logfile,"Player 3 Present: %s\n",HasPlayer3(db)? "Yes" : "No");
	dprintf(logfile,"Player 4 Present: %s\n",HasPlayer4(db)? "Yes" : "No");
	pmsk = 0;
	if (HasPlayer4(db))
		pmsk++;
	if (HasPlayer3(db))
		pmsk++;
	if (HasPlayer2(db))
		pmsk++;
	if (HasPlayer1(db))
		pmsk++;
	if (pmsk)
	{
		dur = (len-1-13)/(4*35*pmsk);
		dprintf(logfile,"Demo runs for %s\n",Duration(dur));
	}
	else
		dprintf(logfile,"Demo time not computable\n");
}
	
void DumpMoves(char *db,int len)
{
	int i,a;
	char act[16];

	dprintf(logfile,"\nMoves by gametic:\n");
	dprintf(logfile," #    ForBackward   LRStrafe  LRTurn  Fire  Push      Switch\n");

	for (i=0;i<NumberTics(db,len);i++)
	{
		a = GetActions(db,i);
		if (a&128)
		{
			if ((a&3)==1)
				strcpy(act,"(Un)Pause");
			else if ((a&3)==2)
			{
				sprintf(act,"Save %d",(a>>2)&7);
			}
			else if (a==128)
			{
				sprintf(act,"Quit");
			}
			else strcpy(act,"Unknown");
		}
		else if (a&4)
		{
			switch((a>>3)&7)
			{
				case 0:
					strcpy(act,"Chainsaw");
					break;
				case 1:
					strcpy(act,"Pistol");
					break;
				case 2:
					strcpy(act,"Shotgun");
					break;
				case 3:
					strcpy(act,"Chaingun");
					break;
				case 4:
					strcpy(act,"Launcher");
					break;
				case 5:
					strcpy(act,"Plasma");
					break;
				case 6:
					strcpy(act,"BFG");
					break;
				case 7:
					strcpy(act,"Dbl Barrel");
					break;
			}
		}
		else strcpy(act," - ");
		dprintf(logfile,"%5d %10d %9d %8d %5s %5s %11s\n",i,GetFBMotion(db,i),GetSLRMotion(db,i),GetTLRMotion(db,i),((a&0x81)==1)? "Fire" : " - ", ((a&0x82)==2)? "Push" : " - ",act);
	}
	dprintf(logfile,"\n");
}

int IsDemoName(char *dnam)
{
	return (strnicmp(dnam,"DEMO",4)==0 && isdigit(dnam[4]) && dnam[5]=='\0')? 1 : 0;
}

void FreeDemStructures()
{
	FreeDir(&pdir);
	FreeDir(&bdir);
}

int demmain(int argc,char **argv)
{
	int i,n,m;
	char *p,mapsel[128],newver[64];
	int cleared=0;
	FILE *tst=NULL,*lst=NULL;
	UBYTE demonum;
	UBYTE levnum=1;

	dprintf(logfile,"\n");

	CopyEnvironmentCLArgs("DEMON",argc,argv);
	if (ParseCommandLine(-1,2))
	{
		FreeCLStrs();
		return 3;
	}

	if (nclfiles!=1 && nclfiles!=2)
	{
		dprintf(logfile,"\nDEMON.EXE by Jim Flynn, Copyright 1996-7 v1.0 -- Freeware --\n\n");
		dprintf(logfile,"From built-in or standalone demo, dump header, motion data\n");
		dprintf(logfile,"or change version number or level\n");
		dprintf(logfile,"\nUsage: DEMON option file{.WAD|.LMP} [DEMO#]\n");
		dprintf(logfile,"\n Options, use SET DEMON=options to set defaults:\n");
		dprintf(logfile," +/-H show demo header or not\n");
		dprintf(logfile," -/+M show demo moves or not\n");
		dprintf(logfile," /Vm.n convert version to m.n\n");
		dprintf(logfile," /W:wadpath convert to version of DEMO1 in wadpath\n");
		dprintf(logfile," /Lnn convert to Mapnn\n");
		dprintf(logfile," /T:nnn Truncate demo by nnn gametics\n");
		dprintf(logfile,"\n See DEMON.TXT for details.\n");

		FreeCLStrs();
		return(1);
	}

	cleared=0;
	strcpy(outname,"");
	ConvertVersion=0;
	ConvertLevel=0;
	for (i=0;i<nclopts;i++)
	{
		if (tolower(cloptions[i][0])=='v')
		{
			strcpy(newver,cloptions[i]);
			ConvertVersion=1;
		}
		else if (tolower(cloptions[i][0])=='l')
		{
			levnum = atoi(cloptions[i]+1);
			ConvertLevel=1;
		}
		else if (tolower(cloptions[i][0])=='h')
		{
			ShowHeader=cloptsense[i];
		}
		else if (tolower(cloptions[i][0])=='m')
		{
			ShowMoves=cloptsense[i];
		}
		else if (tolower(cloptions[i][0])=='w' && cloptions[i][1]==':')
		{
			strcpy(bwadname,cloptions[i]+2);
			FromWad=1;
		}
		else if (tolower(cloptions[i][0])=='t' && cloptions[i][1]==':')
		{
			if (StripPause)
				dprintf(logfile,"/T:%d ignored, already doing /S\n",TruncDemo);
			else
				TruncDemo=atoi(cloptions[i]+2);
		}
		else dprintf(logfile,"Warn: command line option: %s not understood\n",cloptions[i]);
	}

	strcpy(pwadname,clfiles[0]);
	strcpy(mapsel,"");
	if (nclfiles>1 && strnicmp(clfiles[1],"DEMO",4)==0)
	{
		strcpy(mapsel,clfiles[1]);
		SingleMap=atoi(clfiles[1]+4);
		if (SingleMap<1 || SingleMap>4)
		{
			dprintf(logfile,"Only DEMO1, DEMO2, DEMO3, or DEMO4 may be specified\n");
			dprintf(logfile,"\n");
			FreeCLStrs();
			return (3);
		}
	}
	FreeCLStrs();

	if (FromWad)
	{
		AddExt(bwadname,".WAD");

		bst = fopen(bwadname,"rb");
		if (bst)
		{
			fread(&bhdr,sizeof(WadHeader),1,bst);
			ReadDir(&bhdr,&bdir,bst);
			m = IsIn("DEMO1",&bdir);
			if (m>=0)
			{
				fseek(bst,bdir.dirp[m].resptr,SEEK_SET);
				fread(&demonum,1,sizeof(UBYTE),bst);
				demonum -= 100;
				if (demonum>=6 && demonum<=9)
					sprintf(newver,"1.%d",(int)demonum);
				fclose(bst);
				ConvertVersion=1;
			}
			else
			{
				dprintf(logfile,"From wad: %s does not contain a DEMO1, exiting\n",bwadname);
				dprintf(logfile,"\n");
				return 3;
			}
			FreeDir(&bdir);
		}
		else
		{
			dprintf(logfile,"From wad: %s not found, exiting\n",bwadname);
			dprintf(logfile,"\n");
			return 3;
		}
	}

	p=strrchr(pwadname,'.');
	if (p && stricmp(p+1,"WAD")==0)
	{
		pst = (TruncDemo || ConvertVersion || ConvertLevel)? fopen(pwadname,"r+b") : fopen(pwadname,"rb");
		if (pst!=NULL)
		{
			fread(&phdr,sizeof(WadHeader),1,pst);
			ReadDir(&phdr,&pdir,pst);

			for (i=0;i<pdir.nentries;i++)
			{
				if (IsDemoName(pdir.dirp[i].resname) && (!SingleMap || strnicmp(mapsel,pdir.dirp[i].resname,5)==0))
				{
					n = pdir.dirp[i].reslen;
					fseek(pst,pdir.dirp[i].resptr,SEEK_SET);
					demobuff = realloc(demobuff,n);
					fread(demobuff,n,1,pst);

					if (ShowHeader) DumpHeader(demobuff,n,Str8(pdir.dirp[i].resname));
					if (ShowMoves) DumpMoves(demobuff,n);
					if (ConvertVersion)
					{
						if (*demobuff != demonumber(newver))
						{
							*demobuff=demonumber(newver);
							fseek(pst,pdir.dirp[i].resptr,SEEK_SET);
							SafeWrite(demobuff,1,1,pst);
							dprintf(logfile,"Internal %s in %s version changed to %s\n",Str8(pdir.dirp[i].resname),pwadname,newver);
						}
						else dprintf(logfile,"Internal %s in %s version already %s\n",Str8(pdir.dirp[i].resname),pwadname,newver);
					}
					if (ConvertLevel)
					{
						if (demobuff[3] != levnum)
						{
							demobuff[3]=levnum;
							fseek(pst,pdir.dirp[i].resptr+3,SEEK_SET);
							SafeWrite(demobuff+3,1,1,pst);
							dprintf(logfile,"Internal %s in %s level changed to %d\n",Str8(pdir.dirp[i].resname),pwadname,levnum);
						}
						else dprintf(logfile,"Internal %s in %s level already is %d\n",Str8(pdir.dirp[i].resname),pwadname,levnum); 
					}
					if (TruncDemo>0)
					{
						SetNumberTics(&demobuff,NumberTics(demobuff,n)-TruncDemo,&n);
						fseek(pst,pdir.dirp[i].resptr,SEEK_SET);
						SafeWrite(demobuff,n,1,tst);
						dprintf(logfile,"Internal %s in %s truncated by %d gametics\n",Str8(pdir.dirp[i].resname),pwadname,TruncDemo);
					}
					free(demobuff);
					demobuff=NULL;
				}
			}
			fclose(pst);
		}

	}
	else if (p && stricmp(p+1,"LMP")==0)
	{
		tst = (TruncDemo || ConvertVersion || ConvertLevel)? fopen(pwadname,"r+b") : fopen(pwadname,"rb");
		n = filelen(tst);
		if (tst!=NULL && n > 0)
		{
			demobuff = realloc(demobuff,n);
			fread(demobuff,n,1,tst);

			if (ShowHeader) DumpHeader(demobuff,n,pwadname);
			if (ShowMoves) DumpMoves(demobuff,n);
			if (ConvertVersion)
			{
				if (*demobuff != demonumber(newver))
				{
					*demobuff=demonumber(newver);
					fseek(tst,0,SEEK_SET);
					SafeWrite(demobuff,1,1,tst);
					dprintf(logfile,"Demo version of %s changed to %s\n",pwadname,newver);
				}
				else dprintf(logfile,"Demo version of %s already is %s\n",pwadname,newver);
			}
			if (ConvertLevel)
			{
				if (demobuff[3]!=levnum)
				{
					demobuff[3]=levnum;
					fseek(tst,3,SEEK_SET);
					SafeWrite(demobuff+3,1,1,tst);
					dprintf(logfile,"Demo level of %s changed to %d\n",pwadname,levnum);
				}
				else dprintf(logfile,"Demo level of %s already is %d\n",pwadname,levnum);
			}
			if (TruncDemo>0)
			{
				i=NumberTics(demobuff,n);
				if (i>TruncDemo)
				{
					SetNumberTics(&demobuff,i-TruncDemo,&n);
					lst=fopen("tmp_tmp.lmp","wb");
					SafeWrite(demobuff,n,1,lst);
					fclose(lst);

					fclose(tst);
					remove(pwadname);
					rename("tmp_tmp.lmp",pwadname);
					dprintf(logfile,"Demo in %s truncated by %d gametics\n",pwadname,TruncDemo);
					return 0;
				}
				else dprintf(logfile,"Can't delete more gametics than there are\n");
			}
			else if (StripPause)
			{
				RemovePauses(&demobuff,&n);
				lst=fopen("tmp_tmp.lmp","wb");
				SafeWrite(demobuff,n,1,lst);
				fclose(lst);

				fclose(tst);
				remove(pwadname);
				rename("tmp_tmp.lmp",pwadname);
				dprintf(logfile,"Demo in %s stripped of pauses\n",pwadname);
				return 0;
			}
			free(demobuff);
			demobuff=NULL;
			fclose(tst);
		}
	}
	dprintf(logfile,"\n");
	return 0;
}
