#include "patch.h"

void FreeHalfPlane(HalfPlane **hp)
{
	if (!*hp) return;
	else
	{
		HalfPlane *nxt,*h;

		for (h=*hp;h;h=nxt)
		{
			nxt = h->next;
			free(h);
		}
	}
	*hp = NULL;
}

int InHalfPlanes(double x,double y,HalfPlane *hp)
{
	HalfPlane *h;

	for (h=hp;h;h=h->next)
	{
		if ((h->a*x + h->b*y + h->c)*(h->side)<-.1)
			return 0;
	}
  return 1;
}

int IntersectHalfPlanes(HalfPlane *hp1,HalfPlane *hp2,double *x,double *y)
{
	double detx = (hp2->c*hp1->b - hp1->c*hp2->b);
	double dety = (hp2->a*hp1->c - hp1->a*hp2->c);
	double det  = (hp1->a*hp2->b - hp2->a*hp1->b);

	if (fabs(det)>.001)
	{
		*x=detx/det;
		*y=dety/det;
		return 1;
	}
	return 0;
}

static double cx,cy;

int PgnSortByAngle(const void *p1,const void *p2)
{
	double a1,a2;

	a1 = atan2(((int *)p1)[1]-cy,((int *)p1)[0]-cx);
	if (a1<0) a1 += 2*pi;
	a2 = atan2(((int *)p2)[1]-cy,((int *)p2)[0]-cx);
	if (a2<0) a2 += 2*pi;
	return (int)(4096*(a2-a1)); // gives clockwise sort
}

void HPListToPolygon(HalfPlane *hp,int (*pgn)[2],int *npts)
{
	HalfPlane *h1,*h2;
	int i,j;
	double x,y;

	*npts = 0;
	if (!hp) return;

	for (h1=hp;h1;h1=h1->next)
		for (h2=h1->next;h2;h2=h2->next)
		{
			int v = IntersectHalfPlanes(h1,h2,&x,&y);

			if (v) // if lines not parallel
			{
				// insure the vertex is on the right side of all planes
				// then store the point

				if (InHalfPlanes(x,y,hp))
				{
					pgn[*npts][0] = (int)floor(x+.5);
					pgn[(*npts)++][1] = (int)floor(y+.5);
				}
			}
		}

	// now we have a set of points that comprise the vertices of the
	// polygon (in theory anyway). Sort them by angle.

	cx=0;
	cy=0;

	for (i=0;i<*npts;i++)
	{
		cx += pgn[i][0];
		cy += pgn[i][1];
	}

	if (*npts)
	{
		cx /= *npts;
		cy /= *npts;

		qsort(pgn[0],*npts,2*sizeof(int),PgnSortByAngle);
	}

	// remove duplicate vertices
	for (i=0;i<*npts;i++)
	{
		j=i+1;
		while (j<*npts && pgn[i][0]==pgn[j][0] && pgn[i][1]==pgn[j][1])
		{
			j++;
		}

		// delete the redundant entries from i+1 to j-1
		if (j>i+1)
		{
			if (j<*npts)
				memmove(pgn[i+1],pgn[j],2*sizeof(int)*(*npts-j));
			*npts -= j-i-1; // loop limit change!
		}
	}
	if (*npts>1 && pgn[0][0]==pgn[*npts-1][0] && pgn[0][1]==pgn[*npts-1][1])
		(*npts)--;
}


void FrontToBackSubSectors(int nodenum,int viewsub,numlist_t **headsubp)
{

	while (!(nodenum & NF_SUBSECTOR))
	{
		int vx = Vertexes[Segs[Ssectors[viewsub].seg0].from].x;
		int vy = Vertexes[Segs[Ssectors[viewsub].seg0].from].y;
		int side = PointOnNodeSide(vx,vy,Nodes+nodenum);

	  FrontToBackSubSectors(Nodes[nodenum].children[side],viewsub,headsubp);
		nodenum = Nodes[nodenum].children[1^side];
	}

	// pre-pend each subsector encountered onto front-to-back order list
	// note tree traversal order is back-to-front
	{
		numlist_t *newsub = malloc(sizeof(numlist_t));
		newsub->num = nodenum & ~NF_SUBSECTOR;
		newsub->next = *headsubp;
		*headsubp = newsub;
	}
}

// this routine creates for each subsector a halfplane list that
// specifies it and a polygon that is its boundary

static HalfPlane **hphead=NULL;
static int (**subpgn)[2]=NULL;
static int *subnpts=NULL;

static void R_ListSubSectors(unsh nodenum, int level, int *route)
{
	int i;

	// recursively traverse the BSP tree

	if (nodenum & NF_SUBSECTOR)			// if its a leaf - a subsector
	{
		unsh subsec = nodenum & ~NF_SUBSECTOR;
		HalfPlane *hp,*last=NULL;
		unsh ns=Ssectors[subsec].nsegs;
		unsh s0=Ssectors[subsec].seg0;
		
		// add the nodes along the route to the halfplane list for
		// this subsector

		{
			unsh n=NNodes-1;

			for (i=0;!(n & NF_SUBSECTOR) && i<level;i++)
			{
				short x1 = Nodes[n].x;
				short y1 = Nodes[n].y;
				short x2 = Nodes[n].dx+x1;
				short y2 = Nodes[n].dy+y1;
				int g;

				hp = malloc(sizeof(HalfPlane));
				if (last)
					last->next = hp;
				else
					hphead[subsec] = hp;

				hp->a = -(y2-y1);
				hp->b = x2-x1;
				hp->c = x1*y2-x2*y1;
				hp->side = route[i]? SD_LEFT : SD_RIGHT;
				hp->next = NULL;

				g = agcd(hp->a,agcd(hp->b,hp->c));
				if (g)
				{
					hp->a /= g;
					hp->b /= g;
					hp->c /= g;
				}

				last = hp;
				n = Nodes[n].children[route[i]];
			}
		}

		// add the subsector's lines to the halfplane list
		// subsector is on the right side of all of these lines

		for (i=s0;i<s0+ns;i++)
		{
			short x1 = Vertexes[Segs[i].from].x;
			short y1 = Vertexes[Segs[i].from].y;
			short x2 = Vertexes[Segs[i].to].x;
			short y2 = Vertexes[Segs[i].to].y;
			int g;

			hp = malloc(sizeof(HalfPlane));
			if (last)
				last->next = hp;
			else
				hphead[subsec] = hp;

			hp->a = -(y2-y1);
			hp->b = x2-x1;
			hp->c = x1*y2-x2*y1;
			hp->side = SD_RIGHT;
			hp->next = NULL;

			g = agcd(hp->a,agcd(hp->b,hp->c));
			if (g)
			{
				hp->a /= g;
				hp->b /= g;
				hp->c /= g;
			}

			last = hp;
		}

		// Convert the halfplane list to a set of vertices, sorted by angle

		subpgn[subsec] = malloc(NVertexes*sizeof(int)*2);
		HPListToPolygon(hphead[subsec],subpgn[subsec],subnpts+subsec);
		subpgn[subsec] = realloc(subpgn[subsec],subnpts[subsec]*sizeof(int)*2);

		return;
	}

	route[level] = 0;
	R_ListSubSectors(Nodes[nodenum].children[0],level+1,route);

	route[level] = 1;
	R_ListSubSectors(Nodes[nodenum].children[1],level+1,route);
}


