#include "patch.h"

/*======================================================================
//
// NBLDMAIN.C - nodes, blockmap, and reject builder
// 
// ---------------------------------
// Nodes builder - in brief outline
// ---------------------------------
// 
// 1) Initialize, read level
// 
// 2) Create Segs from Linedefs/Sidedefs
// 	 Initial state 0 Nodes, 1 Ssector containing all segs
// 
// 3) Find a seg in the current Ssector, that when extended, has segs from the
//    current Ssector on both sides of it. Optimize for roughly equal numbers
//    on each side. If not possible at all, then the current set of segs
//    is a convex subsector, and needn't be divided further. Return from
// 	 recursion.
// 
// 4) Where the seg's extension intersects other segs, split them
//    in two, adding a vertex. Create a Ssector to hold all segs
// 	 on the left, leave all the segs on the right in the current
// 	 Ssector. Create a node whose right child is the current Ssector
// 	 and whose left child is the new Ssector.
// 
// 5) Recur to 3 with the left child as current node
//    If the left child splits set the new node created here to have
// 	 left child the node created by the recursion.
// 
// 6) Recur to 3 with the right child as current node
//    If the right child splits set the new node created here to have
// 	 right child the node created by the recursion.
// 
======================================================================*/

////////////////////////////////////////////////////////////////////////
//						|
// CreateSegs	|	No parms, returns 1 on fatal error, 0 if ok
//						|
// This routine goes thru the Linedefs and for each Sidedef of a Linedef
// creates a Seg in the same direction as the Sidedef. Segs always go
// clockwise around the sector boundaries. When a seg is created in
// the opposite direction to from->to of the linedef, its dir bit is set.
//
// Initialize the nodes before building them, zero Nodes, one Ssector
// containing all segs created by this routine.
//
// Building will consist of splitting this subsector in half recursively
// until only convex subsectors result
//
////////////////////////////////////////////////////////////////////////

int CreateSegs()
{
	int i;

	// avoid needing to allocate per seg by overallocating

	Segs = malloc(2*NLinedefs*sizeof(Seg));
	NSegs = 0;

	// Go thru the Linedefs - Sidedefs may be packed

	for (i=0;i<NLinedefs;i++)
	{
		// if no first sidedef, return an error

		if (Linedefs[i].sides[0]==-1)
		{
			printf("Fatal: Linedef %d does not have a first side\n",i);
			return 1;
		}

		if (Linedefs[i].sides[1]==-1)
		{
	 		// 1S line case

			// 1st sidedef
			Segs[NSegs].from = Linedefs[i].from;
			Segs[NSegs].to = Linedefs[i].to;
			Segs[NSegs].ang = VertNum2Angle(Linedefs[i].from,Linedefs[i].to);
			Segs[NSegs].ldef = i;
			Segs[NSegs].dir = 0;
			Segs[NSegs++].offs = 0;
		}
		else
		{
			// 2S line case

			// 1st sidedef seg

			Segs[NSegs].from = Linedefs[i].from;
			Segs[NSegs].to = Linedefs[i].to;
			Segs[NSegs].ang = VertNum2Angle(Linedefs[i].from,Linedefs[i].to);
			Segs[NSegs].ldef = i;
			Segs[NSegs].dir = 0;
			Segs[NSegs++].offs = 0;

			// 2nd sidedef seg is flipped

			Segs[NSegs].from = Linedefs[i].to;
			Segs[NSegs].to = Linedefs[i].from;
			Segs[NSegs].ang = VertNum2Angle(Linedefs[i].to,Linedefs[i].from);
			Segs[NSegs].ldef = i;
			Segs[NSegs].dir = 1;
			Segs[NSegs++].offs = 0;
		}
	}

	// Now shrink the overallocated Segs to the correct size

	Segs = realloc(Segs,NSegs*sizeof(Seg));

	// then add one subsector that is a list of all Segs

	AddSsector(NSegs,0);

	// Some debug printout to verify which segs were created

	if (Verbose)
		DumpSegs();

	return 0;		// no errors detected
}

////////////////////////////////////////////////////////////////////////
//
// SegSegCompare
//

SegCmpTypes SegSegCompare(Seg *line,Seg *seg,long *f,long *t)
{
	int x1=X1Of(line),y1=Y1Of(line),x2=X2Of(line),y2=Y2Of(line);
	int w1=X1Of(seg),z1=Y1Of(seg),w2=X2Of(seg),z2=Y2Of(seg);
	int dx1=x2-x1, dy1=y2-y1, dx2=w2-w1, dy2=z2-z1, dx3=w2-x1, dy3=z2-y1;

	// if line intersects seg

	return LineCompare(dx1,dy1,dx2,dy2,dx3,dy3,f,t);
}

////////////////////////////////////////////////////////////////////////
//						  |
// PickNodeLine |	Passed subsector to pick a split for
//						  |	Returns best seg from subsector to split subsector with
//							
// The segs in the 'subsector' are scanned
//
// For each trial splitting seg 
//  count the sets of segs in the subsector that are:
//
//    1) left   - one endpoint to left of line, other left of or on line
//    2) right  - one endpoint to right of line, other right of or on line.
//    3) split  - endpoints on opposite side of split line, or both on line
//    4) effect - seg would be split and is tagged > 900
//
// The cost function for a trial split is:
//
// BalanceWeight * abs(left-right) +		/* how unbalanced */
// SplitWeight   * split +							/* how many split seg */
// EffectWeight  * effect								/* how many 900 tags split */
//
// The weight values are controlled externally to vary the
// the splitting technique.
//
// If all the segs are on one side of every split, the subsector is 
// convex and we return NOSPLIT indicating a real subsector has been
// reached and subdivision can stop here
//
////////////////////////////////////////////////////////////////////////


#define NOSPLIT 0xffff		/* signal for no split found */

int BalanceWeight=1;			// controls how important equal size splits are
int SplitWeight=17;				// controls how important minimizing splits is
int EffectWeight=100;     // controls avoiding 900 line tags

int LargeSplit=10;				// allows all segs to be costed every split

void PickNodeLine(unsh subsecnum,unsh *bestsplit)
{
	unsh ns = Ssectors[subsecnum].nsegs;
	unsh s0 = Ssectors[subsecnum].seg0;
	unsigned cheapest=MAXPOSINT,cost=0;
	int i,j,try,step;

	// default return seg to no split possible

	*bestsplit=NOSPLIT;

	// for each candidate seg in the subsector
	// count up its cost to use as a splitting node

	// if number of possible splitting segs is large
	// try a sampling of segs to speed up early splits
	// idea from John Carmack's IDBSP
	
	step = (LargeSplit && ns>=LargeSplit)? (int)sqrt((double)ns) : 1;
  for (try=0;try<2;try++)
	{
		for (i=s0;i<s0+ns;i+=step)
		{
			Seg *sgi = Segs+i;
			int split=0,left=0,right=0,effect=0;

			// for each seg in subsecnum, sgj, determine its relationship to
			// the candidate splitting seg, sgi.

			for (j=s0;j<s0+ns;j++)
			{
				Seg *sgj = Segs+j;
				long f,t;

				switch(SegSegCompare(sgi,sgj,&f,&t))
				{
					case SEG_LCOINC:							// cost segs to left of sgi
					case SEG_LEFT:
						left+=BalanceWeight;
						break;
					case SEG_RCOINC:							// cost segs to right of sgi
					case SEG_RIGHT:
						right+=BalanceWeight;
						break;
					case SEG_SPLIT:								// cost segs split by sgi
						split+=SplitWeight;					
						if (SegLin(sgj)->tag >= 900)
							effect+=EffectWeight;
						break;
				}

				// if doing a LargeSplit first pass
				// halt the costing if we exceed the best so far
				// balance could improve later, but probably won't
				// idea from Lee Killough's BSP

				cost = abs(left-right)+split+effect;
				if (cost>cheapest && step!=1)
					goto trynextsplit;
			}

			// evaluate split, and keep track of best found

			if (Verbose)
				printf("Split: %4d CH: %6u NS: %4u L:%4d R:%4d S:%4d E: %4d\n",i,cheapest,ns,left,right,split,effect);

			// if any split was found and the cost is less than the best so far
			// remember the new best split, and set the new cheapest cost

			if (left+split>0 && right+split>0 && cost<cheapest)	
			{
				cheapest = cost;
				*bestsplit = i;										
			}
trynextsplit:;
		}
		// if we found a split don't need another try
		// else try again, this time all segs, with step=1

		if (*bestsplit!=NOSPLIT || step==1)
			break;
		step = 1;
	}
}


///////////////////////////////////////////////////////////////////////////////////
//					     |
// ClassifySegs  |  Passed a subsector number, and split seg number
//  				     |	Returns 3 unsh arrays of indices: segs wholly left of
//               |  the split, wholly right of the split, or divided by it
//
// This routine creates three arrays of lineseg indices that
// represent the linesegs on the left of, split by, and to the right of
// the split represented by splitseg. Each seg is classified by which
// side of the split its endpoints lie. 
//
///////////////////////////////////////////////////////////////////////////////////

void ClassifySegs
(
	unsh subsecnum,unsh splitseg,
	unsh *left,int *nleft,
	unsh *right,int *nright,
	unsh *split,int *nsplit
)
{
	unsh ns = Ssectors[subsecnum].nsegs;
	unsh s0 = Ssectors[subsecnum].seg0;
	Seg *sg=Segs+splitseg;
	int i,j;
	SegCmpTypes sc;
	long f,t;

	for (j=s0;j<s0+ns;j++)
	{
		Seg *sgj=Segs+j;

		if (j==splitseg)
			continue;

		// which side of splitseg is sgj?

		sc = SegSegCompare(sg,sgj,&f,&t);

		if (sc==SEG_SPLIT)
			split[(*nsplit)++]=j;												// split
		else if (sc<SEG_SPLIT)
			right[(*nright)++]=j;												// SEG_RIGHT or SEG_RCOINC
		else if (sc>SEG_SPLIT)
			left[(*nleft)++]=j;													// SEG_LEFT or SEG_LCOINC
	}

	// if no splits and one side empty, put split line in that side
	// This is required to avoid HOM's near linedefs referenced to 
	// sectors other than the one they occur in. It guarantees that
	// all segs in one subsector come from the same sector.

	if (!*nsplit && !*nleft)
		left[(*nleft)++]=splitseg;
	else
		right[(*nright)++]=splitseg;

	if (Verbose)
	{
		printf("Split segs: ");
		for (i=0;i<*nsplit;i++)
			printf("%s ",FormatSeg(Segs+split[i]));
		printf("\n");
	}
}

//////////////////////////////////////////////////////////////////////////
//					 |
// SplitSeg  | Passed a segment to split by, a segment to be split
//					 | splits the seg if returns 1 - *side is side new seg is on
//           | seg not split if returns 0 - *side is whether left or right
//
// Split seg2split by extension of splitseg
//
// creates a new Vertex and a new Seg if the seg is split
//
// return true if seg was split
//   side then returns which side new seg is on
//
// return false if seg was not split
//	 side is -1 or -1 if intersection was an endpoint - side seg lies on
//   side is -1 or 1 if seg coincident and opposite or along splitseg resp.
//
//////////////////////////////////////////////////////////////////////////

int SplitSeg(unsh splitseg,unsh seg2split,int *side)
{
	Seg *sps=Segs+splitseg,*s2s=Segs+seg2split;

	long bx1 = X1Of(s2s);
	long by1 = Y1Of(s2s);
	long bx2 = X2Of(s2s);
	long by2 = Y2Of(s2s);

	long ex1 = X1Of(SegLin(s2s));
	long ey1 = Y1Of(SegLin(s2s));
	long ex2 = X2Of(SegLin(s2s));
	long ey2 = Y2Of(SegLin(s2s));

	long bdx = bx2-bx1;
	long bdy = by2-by1;

	long s1,s2;
	
	SegCmpTypes sc = SegSegCompare(sps,s2s,&s1,&s2);

	*side=sc;								// if not split, side is SegCmpTypes

	if (sc==SEG_SPLIT) 			// seg2split is not coincident with splitseg
	{
		unsh newvertex;
		double x,y,x0,y0;
		char *p,*q;
		double dist=SegLen(s2s);
		int oldnv;

		dist = sqrt(bdx*bdx+bdy*bdy);

		if (labs(s1)>=labs(s2))
		{
			x = bx1+(labs(s1)*bdx)/dist;
			y = by1+(labs(s1)*bdy)/dist;
		}
		else
		{
			x = bx2-(labs(s2)*bdx)/dist;
			y = by2-(labs(s2)*bdy)/dist;
		}

		// Add the new Vertex

		oldnv = NVertexes;
		newvertex = AddNewVertex(x0=x,y0=y);

		if (newvertex == s2s->from)
		{
			aprintf("DEBUG: Zero length split: seg %s by %s\n",p=FormatSeg(Segs+seg2split),q=FormatSeg(Segs+splitseg));
			*side = PointSegSideCompare(bx2,by2,sps);
			free(p); free(q);
			return 0;
		}
		else if (newvertex == s2s->to)
		{
			aprintf("DEBUG: Zero length split: seg %s by %s\n",p=FormatSeg(Segs+seg2split),q=FormatSeg(Segs+splitseg));
			*side = PointSegSideCompare(bx1,by1,sps);
			free(p); free(q);
			return 0;
		}

		if (Verbose)
		{
			printf("Adding vertex %d on seg %s\n",newvertex,p=FormatSeg(Segs+seg2split));
			free(p);
		}

		// Splitting a seg:
		//
		// seg2split ---> first seg   second (new) seg
		// -----------------------------------------------
		//	from					from					new vert
		//	to					  new vert			to
		//	ang						ang					  ang
		//  ldef					ldef					ldef
		//	dir						dir						dir
		//	offs					offs					offs+len(first seg)
		//

	  // Change seg2split to be to the new Vertex
		// Correct the offset to the new seg


	 	x=s2s->dir? ex2 : ex1;
	 	y=s2s->dir? ey2 : ey1;

		// Add a new Seg, just like seg2split, except from the new Vertex
		// and a possibly changed offset along the linedef

		AddSeg
		(
			newvertex,
			s2s->to,
			s2s->ang,
			s2s->ldef,
			s2s->dir,
			(unsigned short)Dist(x0,y0,x,y)
		);
	  s2s = Segs+seg2split;	// reset pointers in case Segs moved on realloc
		sps = Segs+splitseg;
	
		s2s->to = newvertex;							 				// make seg2split to new vert

		// return whether the new seg is on the left or right side of splitseg

		*side = s2;																// side of split new seg on

		// if seg is on linedef with >900 tag warn user

		if (SegLin(s2s)->tag>900)
			aprintf("\nWarning: Linedef %d, marked precious, tag = %d was split\n\n",
		        	 s2s->ldef,SegLin(s2s)->tag);

		return 1; 	  														// seg was split
	}

	// seg2split is coincident with splitseg
	// return whether it is in the same or opposite direction

	return 0;																		// seg was not split
}

//////////////////////////////////////////////////////////////////////////
//					 |
// SplitSegs | Passed a segment to split by, and arrays of indices of
//					 | segs left, split by, and to the right of segsplit.
//           | Splits the segs, puts the pieces in the appropriate arrays
//
// Split the segs split by splitseg and put them away in left,right arrays
//
//////////////////////////////////////////////////////////////////////////

void SplitSegs
(
	unsh splitseg,
	unsh *left,int *nleft,
	unsh *right,int *nright,
	unsh *split,int *nsplit
)
{
	int j,side;

	for (j=0;j<(*nsplit);j++)
	{
		// split seg split[j] with splitseg

		if (split[j]==splitseg)
		{
			right[(*nright)++]=split[j];
		}
		else if (SplitSeg(splitseg,split[j],&side)) 		// seg was split
		{
			// put each half in the appropriate left/right lists

			left[(*nleft)++] = (side<0)? split[j] : NSegs-1;
			right[(*nright)++] = (side<0)? NSegs-1 : split[j];
		}
		else																			// seg was not split
		{
			// put the unsplit seg in the left/right list it belongs in

			if (side>0)
				left[(*nleft)++] = split[j];
			else if (side<0)
				right[(*nright)++] = split[j];
			else
				printf("DEBUG: Unassigned seg\n");
		}
	}
}

//////////////////////////////////////////////////////////////////////////
//
//             |  Create a lookup table for a predicate on segs in
// initisleft  |  the current subsector for whether they are left
//             |  or right of the current split
// 
// freeisleft  |  Deallocate the lookup table
//						 |
//
//////////////////////////////////////////////////////////////////////////

UBYTE *isleft=NULL;

void initisleft(unsh s0,unsh ns,unsh *left,int nleft)
{
	int i;

	isleft = realloc(isleft,ns);
	memset(isleft,0,ns); 				// set all to right

	for (i=0;i<nleft;i++)
		isleft[left[i]-s0]=1;			// make the left segs left
}

// kill the lookuptable

void freeisleft()
{
	free(isleft);
	isleft=NULL;
}

////////////////////////////////////////////////////////////////////////////
//
//           | Recursively divides a 'subsector' containing segs into
// SplitNode | smaller subsectors until only convex subsectors remain.
//					 | 
//					 | Each recursion one Node and one new Ssector are created
//
// PickNodeLine is called and if it returns NOSPLIT, the subsector
// is convex, we return.
//
// if PickNodeLine returns a seg number to use as node line then
// we perform splits on the segs and sort the enlarged set into
// left of node seg and right of node seg sets.
//
// We create a new Node and a new Ssector. The right of node segs
// are put in the old Ssector we were passed and the left of node
// segs are put in the new Ssector we created. We add the new Ssector
// as the left child of the new Node and the old Ssector as its right
// child. We then fill out its data from the node seg and two sets of segs.
// 
// We then recur to the left child
// On return we recur to the right child
// return to original caller occurs as the tree is completed
//

int SplitNode(unsh nodenum)
{
	unsh splitseg,subsecnum;
	unsh *left,*right,*split,ns,s0;
	int nleft,nright,nsplit,a,b;
	int oldnsegs,i,j,k,N,nnodes;
	static int count=0;

	subsecnum = nodenum & ~NF_SUBSECTOR;

	// Look thru the segs listed by this subsector for one that splits
	// those segs into roughly equal parts, without too many splits

	PickNodeLine(subsecnum,&splitseg);

	if (Verbose)
		if (splitseg==NOSPLIT)
			printf("No split found for subsector %5u\n",subsecnum);
		else
		{
			char *p;
			printf("Splitting subsector %5u with seg %s\n",subsecnum,p=FormatSeg(Segs+splitseg));
			free(p);
		}

	// If PickNodeLine can't find any splits, all the regions are convex
	//
	// In the presence of unclosed sectors though, we may have subsectors
	// including segs from more than one sector.
	//
	// If this is the case let ClassifySegs move the partition seg into
	// its own subsector
	//

	ns = Ssectors[subsecnum].nsegs;
	s0 = Ssectors[subsecnum].seg0;

	if (splitseg==NOSPLIT)
	{
		int s = Sidedefs[Linedefs[Segs[s0].ldef].sides[Segs[s0].dir]].sect;
		for (i=s0;i<s0+ns;i++)
			if (Sidedefs[Linedefs[Segs[i].ldef].sides[Segs[i].dir]].sect!=s)
			{
				splitseg = i;
				break;
			}
		if (i==s0+ns)
			return MAXNEGINT;
	}

	// Now we determine which segs are to the left, right or split by
	// the extension of the seg returned by PickNodeLine

	left = malloc(ns*sizeof(unsh)); nleft = 0;
	right = malloc(ns*sizeof(unsh)); nright = 0;
	split = malloc(ns*sizeof(unsh)); nsplit = 0;

	ClassifySegs(subsecnum,splitseg,left,&nleft,right,&nright,split,&nsplit);

	// Actually split the segs the node split divides
	// All the segs, including new ones are in left or right after this call

	oldnsegs = NSegs;
	SplitSegs(splitseg,left,&nleft,right,&nright,split,&nsplit);

	if (nleft+nright != ns+NSegs-oldnsegs)
		aprintf("DEBUG: Seg lost in sort %d < %d\n",nleft+nright,ns+NSegs-oldnsegs);

	// if in fact, nothing split, check to make sure all segs refer
	// to the same sector. If so, we've got a subsector.

	if (nright==0 || nleft==0)
		return MAXNEGINT;


	// Adjust the segs so that the old and new segs are contiguous
	// The old segs are between s0 and s0+ns-1. The new segs are
	// between oldnsegs and NSegs-1. We move the new segs to just
	// above the old segs.
	//
	//				s0			 s0+ns				oldnsegs	 NSegs
	//				|				 |						|					 |
	//   sssssSSSSSSSSSsssssssssssssNNNNNNNNNNN
 	//														<--->	swap

	if (oldnsegs>s0+ns && NSegs>oldnsegs)
	{
		Seg *temp = malloc(NSegs*sizeof(Seg));
		memmove(temp, Segs+s0+ns, (oldnsegs-s0-ns)*sizeof(Seg));
		memmove(Segs+s0+ns, Segs+oldnsegs, (NSegs-oldnsegs)*sizeof(Seg));
		memmove(Segs+s0+ns+NSegs-oldnsegs, temp, (oldnsegs-s0-ns)*sizeof(Seg));
		free(temp);
	}

	// change all seg references moved

	for (i=0;i<NSsectors;i++)
		if (Ssectors[i].seg0>=s0+ns && Ssectors[i].seg0<oldnsegs)
			Ssectors[i].seg0 += NSegs-oldnsegs;
		else if (Ssectors[i].seg0>=oldnsegs)
			Ssectors[i].seg0 -= oldnsegs-s0-ns;

	for (i=0;i<nright;i++)
		if (right[i]>=s0+ns && right[i]<oldnsegs)
			right[i] += NSegs-oldnsegs;
		else if (right[i]>=oldnsegs)
			right[i] -= oldnsegs-s0-ns;

	for (i=0;i<nleft;i++)
		if (left[i]>=s0+ns && left[i]<oldnsegs)
			left[i] += NSegs-oldnsegs;
		else if (left[i]>=oldnsegs)
			left[i] -= oldnsegs-s0-ns;

	if (splitseg>=s0+ns && splitseg<oldnsegs)
		splitseg += NSegs-oldnsegs;
	else if (splitseg>=oldnsegs)
		splitseg -= oldnsegs-s0-ns;

	// now that the segs are between s0 and s0+ns+(NSegs-oldnsegs)-1, 
	// sort them by whether they are on left or right of splitseg
	// when done, all the left segs in that range should be listed
	// before all the right segs. The following sort is linear in time.

	// construct predicate "isleft" to return which side of split
	// each seg in this subsector lies

	initisleft(s0,ns+NSegs-oldnsegs,left,nleft);

	// scanning, swap any right seg followed by a left seg 
	// result is all left segs on left, all right segs on right

	i = j = 0;
	N = ns+NSegs-oldnsegs;
	while (j<N)
	{
		// find the first right seg

		while (i<N && isleft[i]) 
			i++;

		// find a left seg after it

		j=i+1;
		if (j<N)
		{
			while (j<N && !isleft[j])
				j++;

			// exchange the left and right segs, one less out of place

			if (j<N)
			{
				Seg t = Segs[s0+i];
				Segs[s0+i]=Segs[s0+j];	// swap the segs
				Segs[s0+j]=t;

				k=isleft[i];						// update "isleft" predicate
				isleft[i]=isleft[j];
				isleft[j]=k;

				if (splitseg==s0+j)			// and splitseg if it moved
					splitseg=s0+i;
				else if (splitseg==s0+i)
					splitseg=s0+j;
			}

			// look for the next right seg after this one
			i++;
		}
	}
	freeisleft();

	//
	// Split is complete
	//
	// Now we create a new Node for this seg split, and a new Ssector
	// The old Ssector and the new one are the children of the new node
	//
	// The old subsector (subsecnum) contains the right segs
	// The new subsector contains the segs to the left of the split
	//
	// Note neither of these subsectors are necessarily convex yet until
	// a split fails to be possible for them
	//

	AddSsector(nleft,s0);		// by sort above left segs are s0 thru s0+nleft-1

  // We modify the current subsector to only hold the right segs

  Ssectors[subsecnum].nsegs = nright;		// right segs are s0+nleft
	Ssectors[subsecnum].seg0 = s0+nleft;	// thru s0+nleft+nright-1

	// Add the new node

	AddNode
	(
	 	Vertexes[Segs[splitseg].from].x,
	 	Vertexes[Segs[splitseg].from].y,
	 	Vertexes[Segs[splitseg].to].x - Vertexes[Segs[splitseg].from].x,
	 	Vertexes[Segs[splitseg].to].y - Vertexes[Segs[splitseg].from].y,
		MAXNEGSH,MAXPOSSH,MAXPOSSH,MAXNEGSH,
		MAXNEGSH,MAXPOSSH,MAXPOSSH,MAXNEGSH,
		subsecnum|NF_SUBSECTOR,(NSsectors-1)|NF_SUBSECTOR
	);

  if (((++count)&127)==0)
	  fprintf(stderr,"Nodes: %d\r",NNodes);  // progress

  // Calculate the node bounding boxes

	{
		Node *n = Nodes+NNodes-1;
  	unsh lns = Ssectors[NSsectors-1].nsegs;
		unsh ls0 = Ssectors[NSsectors-1].seg0;
		unsh rns = Ssectors[subsecnum].nsegs;
		unsh rs0 = Ssectors[subsecnum].seg0;

		// left bounding box

  	for (i=ls0;i < ls0+lns;i++)
		{
			Seg *sg = Segs + i;
			Vertex *from = Vertexes + sg->from;
			Vertex *to = Vertexes + sg->to;

			if (from->x < n->lmnx)
				n->lmnx = from->x;
			if (from->x > n->lmxx)
				n->lmxx = from->x;

			if (to->x < n->lmnx)
				n->lmnx = to->x;
			if (to->x > n->lmxx)
				n->lmxx = to->x;

			if (from->y < n->lmny)
				n->lmny = from->y;
			if (from->y > n->lmxy)
				n->lmxy = from->y;

			if (to->y < n->lmny)
				n->lmny = to->y;
			if (to->y > n->lmxy)
				n->lmxy = to->y;
		}

		// right bounding box

  	for (i=rs0;i < rs0+rns;i++)
		{
			Seg *sg = Segs + i;
			Vertex *from = Vertexes + sg->from;
			Vertex *to = Vertexes + sg->to;

			if (from->x < n->rmnx)
				n->rmnx = from->x;
			if (from->x > n->rmxx)
				n->rmxx = from->x;

			if (to->x < n->rmnx)
				n->rmnx = to->x;
			if (to->x > n->rmxx)
				n->rmxx = to->x;

			if (from->y < n->rmny)
				n->rmny = from->y;
			if (from->y > n->rmxy)
				n->rmxy = from->y;

			if (to->y < n->rmny)
				n->rmny = to->y;
			if (to->y > n->rmxy)
				n->rmxy = to->y;
		}

	}
	// free the count arrays

	free(left);
	free(right);
	free(split);

	if (Verbose)
	{
		for (i=0;i<NSsectors;i++)
			printf("Ss%d: %u %u  ",i,Ssectors[i].nsegs,Ssectors[i].seg0);
		printf("\n");
	}

	// Now recur to the two new children of the new node
	// If they split
	//  replace the subsector number with the node numbers they create

	nnodes = NNodes;
	a = SplitNode(Nodes[nnodes-1].children[1]);
	if (a!=MAXNEGINT)
		Nodes[nnodes-1].children[1] = a;

	b = SplitNode(Nodes[nnodes-1].children[0]);
	if (b!=MAXNEGINT)
		Nodes[nnodes-1].children[0] = b;

	return nnodes-1;
}


int cmpssidx(const void *i1,const void *i2)
{
	return Ssectors[*((int *)i1)].seg0 - Ssectors[*((int *)i2)].seg0;
}

int cmpssec(const void *i1,const void *i2)
{
	return ((Ssector *)i1)->seg0 - ((Ssector *)i2)->seg0;
}

// Reverse the nodes, as the engine expects the root at the end

void ReverseNodes()
{
	int i,j,k;
	int *ssidx;

	// flip the nodes upside down

	for (i=0;i<NNodes/2;i++)
	{
		Node n = Nodes[i];
		Nodes[i] = Nodes[NNodes-1-i];
		Nodes[NNodes-1-i] = n;
	}

	// flip the node references too

	for (i=0;i<NNodes;i++)
		for (j=0;j<2;j++)
			if (!(Nodes[i].children[j]&NF_SUBSECTOR))
				Nodes[i].children[j] = NNodes-1-Nodes[i].children[j];

	// sort the subsectors by their first seg thru an index array
	ssidx = malloc(NSsectors*sizeof(int));
	for (i=0;i<NSsectors;i++)
		ssidx[i] = i;
	qsort(ssidx,NSsectors,sizeof(int),cmpssidx);
	qsort(Ssectors,NSsectors,sizeof(Ssector),cmpssec);

	// rereference the ssectors in the nodes
	for (i=0;i<NNodes;i++)
		for (j=0;j<2;j++)
			if ((Nodes[i].children[j]&NF_SUBSECTOR))
				for (k=0;k<NSsectors;k++)
					if (ssidx[k]==(Nodes[i].children[j]&~NF_SUBSECTOR))
					{
				 		Nodes[i].children[j] = k | NF_SUBSECTOR;
						break;
					}
	free(ssidx);


}

void PrintNodes(unsigned short nodenum,int level)
{
	int i,j;

	if (!NNodes) return;

	for (i=0;i<level;i++)
		printf("  ");

	if (nodenum&NF_SUBSECTOR)
	{
		printf("Subsec %d | ",i=nodenum&~NF_SUBSECTOR);
		for (j=0;j<Ssectors[i].nsegs;j++)
		{
			printf("%d:%d->%d ",
							Ssectors[i].seg0+j,
							Segs[Ssectors[i].seg0+j].from,
							Segs[Ssectors[i].seg0+j].to);
		}
		printf("\n");
		return;
	}
	printf("Node   %d\n",nodenum);

	PrintNodes(Nodes[nodenum].children[1],level+1);
	PrintNodes(Nodes[nodenum].children[0],level+1);
}

static int maxlevel=0;
static int lopsided=0;

void NodeCounter(unsigned short nodenum,int level,int *maxlev,int *lops)
{
	int sub0,sub1;

	if (level > *maxlev)
		*maxlev = level;

	if (!NNodes || nodenum&NF_SUBSECTOR)
		return;

	sub0 = Nodes[nodenum].children[0]&NF_SUBSECTOR;
	sub1 = Nodes[nodenum].children[1]&NF_SUBSECTOR;

	if (sub0!=sub1)
		(*lops)++;

	NodeCounter(Nodes[nodenum].children[0],level+1,maxlev,lops);
	NodeCounter(Nodes[nodenum].children[1],level+1,maxlev,lops);
}

void CountNodes(unsigned short nodenum,int *maxlev,int *lops)
{
	*maxlev = *lops = 0;

	NodeCounter(nodenum,0,maxlev,lops);
}

int nbldmain(int argc,char **argv)
{
	char mapsel[256],mapname[256];
	int i,ZeroReject=1,OverWrite=0,DoReject=0;
	int time0=0,time1=0,time2=0,time3=0,time4=0,time5=0,time6=0;
	int nlevs=0;

	time0 = clock();

 	Verbose=0;		

	InitCLStrs();

	CopyEnvironmentCLArgs("NBLD",argc,argv);
	if (ParseCommandLine(-1,-1))
	{
		dprintf(logfile,"Error parsing command line\n");
		FreeCLStrs();
		return 3;
	}

	if (nclfiles<1)
	{
		dprintf(logfile,"\nNBLD v1.1 by Jim Flynn, Copyright 1998  -- Freeware --\n");
		dprintf(logfile,"\nUsage: NBLD [option] wadname[.wad] [MAPnn] [/o:outname[.wad]]\n");
		dprintf(logfile,"Build the nodes and blockmap for a wad\n");
		dprintf(logfile,"\n Options, use SET NBLD=options to set defaults:\n");
		dprintf(logfile,"\n  {+/-}V output lots of information while building [default=OFF]\n");
		dprintf(logfile,"  /O:wadname set name of output wad [default=TMP.WAD]\n");
		dprintf(logfile,"  /O overwrite input wad - use at own risk [default=TMP.WAD]\n");
		dprintf(logfile,"\n  /T:nnn use optimization on splits larger than nnn [default=10,0=OFF]\n");
		dprintf(logfile,"\n  /B:nnn set importance of balancing the tree [default=1]\n");
		dprintf(logfile,"  /S:nnn set importance of avoiding splits [default=17]\n");
		dprintf(logfile,"  /E:nnn set importance of not splitting lines tagged >900 [default=100]\n");
		dprintf(logfile,"\n  -R preserve reject [default: zero reject]\n");
						 
		FreeCLStrs();
		return(1);
	}
	else fprintf(stderr,"\n");

	if (nclfiles<1)
	{
		FreeCLStrs();
		exit(1);
	}

	strcpy(outname,"");
	for (i=0;i<nclopts;i++)
	{
		if (tolower(cloptions[i][0])=='v')
		{
			Verbose = cloptsense[i];
		}
		else if (tolower(cloptions[i][0])=='r')
		{
			if (cloptsense[i])
			{
				DoReject=1;
			  ZeroReject=0;
			}
			else
			{
				ZeroReject=0;
			  DoReject=0;
			}
		}
		else if (tolower(cloptions[i][0])=='t' && cloptions[i][1]==':')
		{
			LargeSplit = atoi(cloptions[i]+2);
		}
		else if (tolower(cloptions[i][0])=='b' && cloptions[i][1]==':')
		{
			BalanceWeight = atoi(cloptions[i]+2);
		}
		else if (tolower(cloptions[i][0])=='s' && cloptions[i][1]==':')
		{
			SplitWeight = atoi(cloptions[i]+2);
		}
		else if (tolower(cloptions[i][0])=='e' && cloptions[i][1]==':')
		{
			EffectWeight = atoi(cloptions[i]+2);
		}
		else if (tolower(cloptions[i][0])=='o' && cloptions[i][1]==':')
		{
			strcpy(outname,cloptions[i]+2);
			AddExt(outname,".WAD");
		}
		else if (tolower(cloptions[i][0])=='o')
		{
			OverWrite = cloptsense[i];
		}
	}

	strcpy(pwadname,clfiles[0]);
	AddExt(pwadname,".WAD");

	if (!*outname)
		if (OverWrite)
			strcpy(outname,pwadname);
		else
			strcpy(outname,"TMP.WAD");

	strcpy(mapsel,"");
	if (nclfiles>1)
		strcpy(mapsel,clfiles[1]);

	FreeCLStrs();

	pst = fopen(pwadname,"rb");
	if (pst)
	{
		int d;

		fread(&phdr,sizeof(WadHeader),1,pst);
		InitDir(&pdir);
		ReadDir(&phdr,&pdir,pst);

		aprintf("\nNBLD v1.0 by Jim Flynn, Copyright 1998  -- Freeware --\n");

		nlevs = 0;
		for (d=0;d<pdir.nentries;d++)
		{
			strcpy(mapname,Str8(pdir.dirp[d].resname));
			if (IsD2Map(mapname) || IsD1Map(mapname))
			{
				if (!*mapsel || stricmp(mapsel,mapname)==0)
				{
					InitLevelData();
					if (!ReadLevelData(mapname,&pdir,pst,11))
					{
						int oldnsegs;

						nlevs++;

						// Rid the wad of unused vertices
						// this assumes they follow all vertices used by linedefs

						for (i=0;i<NVertexes;i++)
							if (!VertUsed(i))
							{
								NVertexes = i;
								Vertexes = realloc(Vertexes,NVertexes*sizeof(Vertex));
								break;
							}

						time1 = clock();

						// initialize the node resources to empty

						free(Segs); Segs=NULL; NSegs=0;
						free(Ssectors); Ssectors=NULL; NSsectors=0;
						free(Nodes); Nodes=NULL; NNodes=0;

						// Create the initial Segs from the Linedefs/Sidedefs

						CreateSegs();
						oldnsegs = NSegs-1;

						aprintf("\nBuilding %s\n",mapname);
						aprintf("Created %5d Segs from %5d Linedefs\n",NSegs,i);
						aprintf("Weights: Balance=%d, Splits=%d, Effects=%d, Threshold=%d\n\n",
				       			BalanceWeight,SplitWeight,EffectWeight,LargeSplit);

						// call pointing to current 'node', subsector 0, containing all segs
						// returns with node tree created, all subsectors convex

						SplitNode(0|NF_SUBSECTOR);

						// Reverse the Nodes and Node references in the tree

						ReverseNodes();

						// Create a Blockmap

						time2 = clock();
						aprintf("Nodes built: %5d Subsectors: %5d Segs: %5d Elapsed time: %3.2f\n",NNodes,NSsectors,NSegs,(time2-(double)time1)/CLOCKS_PER_SEC);

						BuildBlockmap();

						// Create a Reject (for now, zeroed)

						time3 = clock();
						aprintf("\nBlockmap built: %5d bytes                      Elapsed time: %3.2f\n",NBlockmap,(time3-(double)time2)/CLOCKS_PER_SEC);

						if (DoReject)
						{
							BuildReject();
							time4 = clock();
							aprintf("\nReject built.                                    Elapsed time: %3.2f\n",(time4-(double)time3)/CLOCKS_PER_SEC);
						}
						else if (ZeroReject)
						{
							BuildZeroReject();
							time4 = clock();
							aprintf("\nReject zeroed.                                   Elapsed time: %3.2f\n",(time4-(double)time3)/CLOCKS_PER_SEC);
						}
						else
						{
							time4 = clock();
							aprintf("\nReject preserved.                                Elapsed time: %3.3f\n",(time4-(double)time3)/CLOCKS_PER_SEC);
						}

						// write the output level, just the 11 classic resources
						// to a temporary file

						TackLevelToWad("__nlev__.tmp",mapname);

						CountNodes(NNodes-1,&maxlevel,&lopsided);
						if (Verbose)
						{
							PrintNodes(NNodes-1,0);
							DumpNodes();
							DumpSsectors();
							DumpSegs();
						}

						time5 = clock();

						aprintf("\n%s in %s built, time: %3.2f\n",mapname,pwadname,(time5-(double)time1)/CLOCKS_PER_SEC);
						aprintf("Depth: %d  Lopsidedness: %d%%  Splits: %d%%\n\n",maxlevel,(int)(lopsided*100.0/NNodes),(int)((NSegs-oldnsegs)*100.0/NSegs));

						FreeLevelData();
					}
					else
						printf("Can't find %s in %s\n",mapname,pwadname);
				}
			}
		}

		if (WriteTotalWad(outname,pst,&phdr,&pdir,"__nlev__.tmp"))
			aprintf("Cannot open __nbld__.tmp for output\n");

		time6 = clock();
	}
	else
		printf("Can't open %s\n",pwadname);


	aprintf("%2d levels of %s built in %s. Total elapsed time: %3.3f\n",
		      nlevs,pwadname,outname,(time6-(double)time0)/CLOCKS_PER_SEC);

	return 0;
}
