#include "patch.h"

/************************************************************
 PARSECP.C
			 /-------------------------\
			||  CUTPASTE file parsing  ||
			 \-------------------------/
									
************************************************************/

typedef enum
{
	CK_NONE=-1,
	CK_ELSE,
	CK_ELSEIF,
	CK_ENDIF,
	CK_IF,
	CK_SET,
	CK_OPTION,
	CK_CMAP,
} cpkeyidx;

char *cpkeys[]=
{
	".ELSE",
	".ELSEIF",
	".ENDIF",
	".IF",
	".SET",
	".OPTION",
	".COLORMAP"
};
int ncpkeys = sizeof(cpkeys)/sizeof(cpkeys[0]);

extern void DoCpCommandLineOpts();

// Parse the cpname and numeric info in the line
// Done after comment, sign, and path removal

void GetCpInfo(char *line,char *cpname,int maxcpname,int *nnums,char ***nums)
{
	strcpy(cpname,"");			// default to empty and zero

	Tokenize(line," \t",nums,nnums);

	if (*nnums>0)
	{
		strcpy(cpname,(*nums)[0]);
		(*nums)++; (*nnums)--;
	}
	if (*nnums>7) ParseError=ExtraData;
}

//
// Abstract the entities from a make line
// Syntax description:
//
//
// [*][ ]resname [([ ]pathname[ ])] [int1 [int2]][[ ]=[ ]signtext][;comment]
//
//     or
//
// [ ][;comment]\n
//
//
// [parm] represents an optional parm
// [ ] represents 0 or more whitespaces
//
// Syntax summary. Each line is a blank line, a comment, or a data line
//
// Blank lines and comment lines are ignored
// 
// For data lines, the inline comment is defined by the first
// semicolon in the line not preceded by a literalizing backslash, and this
// semicolon and all following characters in the line are removed before
// further parsing.
//
// For data lines, if a ( occurs it must be followed by a ), and 
// the trimmed text between is a path, or filename. A path for a 
// data line describes where the resource for that line is read,
// it may be a directory, or a full file name, either relative to
// the current directory, or a full path.
//
// A data line may start with an asterisk, then after an optional whitespace
// a required resource name followed by two to seven integers, whitespace
// delimited. Then may follow optional whitespace and an equal sign. If present
// the text from the equal sign to the end of line, or start of inline comment
// is a graphic sign described by the text.
//
// A data line may define a sign, by inserting the symbol = 
//
// Also for data lines any occurrence of = outside a comment
// introduces a sign specification which terminates with end of string or
// start of comment. This sign text is removed from the string if present
// after comment removal, and stored to the sign buffer passed.
//
// Simple rules
// 2) A blank line or line starting with whitespace or ; is ignored
// 3) The inline comment begins with the first ; not preceeded by a \
// 4) The inline comment ends at end of line or end of file
// 5) A data line begins with *, or the first letter of the resource name
// 6) Whitespace between * and the resource name is optional
// 7) If the first parentheses in a data line is balanced the text
//    between is the path to the resource named.

void ParseCp
(
	char *line,			// the line we are parsing
	int lineno,			// the line number we are parsing
	int *star, 			// returns if line begins with an asterisk
	char *res,			// returns name of the resource
	char *path,			// returns path to physical file, if any
	char *sign,			// returns sign info, if any
	int *nnums,		  // returns returns number of numbers in array
	char ***nums,			// returns array of numbers after resource name
	int *cond				// returns whether conditional compilation is on/off
)
{
	int kwd;

	ParseError = NoError;

	strcpy(res,"");		// default all return parameters to blank or zero
	strcpy(path,"");    
	strcpy(sign,"");
	*star=0;

	StripComment(line); // remove inline-comment, delete full line comment
	if (line)
	{
		if (*line)
			if (!*star && *line=='.')
			{
				GetKeyword(line,cpkeys,ncpkeys,&kwd);
				switch (kwd)
				{
					case CK_ELSE:
						ParseELSE(line);
						break;
					case CK_ELSEIF:
						ParseELSEIF(line);
						break;
					case CK_ENDIF:
						ParseENDIF(line);
						break;
					case CK_IF:
						ParseIF(line);
						break;
					case CK_SET:
						ParseSET(line);
						break;
					case CK_OPTION:
						ParseOPTION(line);
						DoCpCommandLineOpts();
						break;
					case CK_CMAP:
						ParseCMAP(line);
						break;
					default:
						ParseError = UnrecognizedKeyword;
						break;
				}
			}
			else
			{
				strcpy(res,"");

				if (*line) ExtractSign(line,sign,4096);				// pull out sign
				if (ParseError==SignOverflow)
					dprintf(logfile,"WARNING: Sign overflow in line #%d, truncating\n",lineno);

				// Keep track of largest ParseError since MaxParseError last reset
				if (ParseError > MaxParseError)
					MaxParseError = ParseError;

    		if (*line) ExtractPath(line,path,256);				// pull out path
				if (ParseError==PathOverflow)
					dprintf(logfile,"ERROR: Pathname overflow in line #%d\n",lineno);
				if (ParseError==UnbalParen)
					dprintf(logfile,"ERROR: Unbalanced parentheses in line #%d\n",lineno);

				// Keep track of largest ParseError since MaxParseError last reset
				if (ParseError > MaxParseError)
					MaxParseError = ParseError;

				if (*line) GetCpInfo(line,res,9,nnums,nums);	// recover CPNAM n1... n7
			}

		if (ParseError==ResnameOverflow)
			dprintf(logfile,"ERROR: Resource name longer than 8 in line #%d\n",lineno);
		else if (ParseError==ExtraData)
			dprintf(logfile,"WARNING: extra data following resource info in line #%d\n",lineno);
		else if (*line && *line!='.' && !*res)
		{
			ParseError = NoInfo;
			dprintf(logfile,"ERROR: Missing resource information in line #%d\n",lineno);
		}

		// Keep track of largest ParseError since MaxParseError last reset
		if (ParseError > MaxParseError)
			MaxParseError = ParseError;
	}

	// Delete spaces starting and ending res and path, leave sign alone
	Trim(res);
	Trim(path);
	*cond = ifcondition();
}

