#include "patch.h"

/************************************************************
 PICTURE.C
			 /-------------------------\
			||  DOOM Picture support   ||
			 \-------------------------/
									
************************************************************/

//
// Overlay a subrectangle of the bitmap bpmap over another bitmap, out.
// Subrectangle has origin (xorg,yorg) in bmap and is xsiz by ysiz
// Upper left corner of subrectangle overlaid at (xoff,yoff) in out.
// 

void OverlayBitmaps(Bitmap *out,Bitmap *pbmap,int xoff,int yoff,int light,int cmap,int tmap,int xorg,int yorg,int xsiz,int ysiz)
{
	int i,j;

	// Move the bytes in rectangle (xorg,yorg)-(xorg+xsiz-1,yorg+ysiz-1) of pbmap
	// to the position in out such that the upper left corner of pbmap is at
	// (xoff,yoff) in out. Clip the move to the physical boundaries of out and
	// pbmap. If a byte is xparent in value, do not transfer it. If cmap is
	// non-negative translate the bytes thru a colormap indexed by light/8.
	// If tmap is non-negative mix the input and output pixels according to
	// a translucency table selected.

	for (i=xorg;i<xorg+xsiz;i++)
	{
		int dx=xoff+i-xorg;
		if (dx>=0 && dx<out->logw)
		{
			int li = i%pbmap->realw;
			for (j=yorg;j<yorg+ysiz;j++)
			{
				int dy=yoff+j-yorg;
				int lj=j%pbmap->logh;
				UBYTE pxl=pbmap->bmp[pbmap->logw*lj+li];

				if (dy>=0 && dy<out->logh && pxl!=xparent)
				{
					if (cmap>=0 && cmap<ncmaps && 0<=light && light<256)
						pxl = cmaps[cmap].cmap[31-(light>>3)][pxl];
					if (tmap>=0 && tmap<ntmaps)
					{
						int opxl = out->bmp[out->logw*dy+dx];

						out->bmp[out->logw*dy+dx] = opxl==xparent?
							pxl : tmaps[tmap].tmap[opxl][pxl];
					}
					else out->bmp[out->logw*dy+dx] = pxl;
				}
			}
		}
	}
}

// Overlay a Doom picture into a Bitmap using its internal
// offsets and a specified offset

void OverlayPicture(Bitmap *out,PicData *p,int xo,int yo)
{
	int i,j;
	UBYTE rown,*postp;
	int npix;

	for (i=0;i<p->pwid;i++)
	{
		if (i+xo<0) continue;		// clip to the x boundaries of the output
		if (i+xo>=out->logw) break;

		// do the column, start by making a pointer to it

		postp = p->posts + p->coloffs[i] - 4*sizeof(short) - p->pwid*sizeof(int);

		rown=*postp++;		// get the row offset or column terminator
		while (rown!=255)	// while not a column terminator
		{
			// get the number of pixels in this post
			// if pic height > 254 and post lengths are zeroed use the tall
			// single post column format. Else follow the post length data.

			npix = *postp++;
			if (p->phgt>254 && !npix)
			{
				npix = p->phgt;
				rown = 0;
			}

			// move the pixels in the post

			postp++;  // skip zero byte
			for (j=0;j<npix;j++,postp++)
				if (rown+j+yo>=0 && rown+j+yo<out->logh)	// clip output y boundaries
					out->bmp[out->logw*(rown+j+yo)+i+xo] = Xlation? 
						  XlateColor(*postp,XlateTextColor,Brightness) : *postp;
			postp++; // skip zero byte

			// get next post's row offset or column terminator

			rown=*postp++;  
		}
	}
}

// Convert a Bitmaps bytes from one palette to another

void ConvertBMP(Bitmap *out,RGBQUAD *thepal,RGBQUAD *pal)
{
	int lookup[256];
	int i,j;

	MatchPalettes(thepal,pal,lookup);
	for (i=0;i<out->logh;i++)
		for (j=0;j<out->logw;j++)
			out->bmp[out->logw*i+j] = lookup[out->bmp[out->logw*i+j]];
}

// Write a Bitmap to a .BMP file

void WriteBMP(Bitmap *out,char *rbmpname,RGBQUAD *pal)
{
	int i;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	int fhsiz,ihsiz;
	FILE *st;
	char bmpname[256],tbmpname[256];
	char *p=NULL;

	strcpy(tbmpname,rbmpname);
	strupr(tbmpname);
	p = strstr(tbmpname,"VILE\\");
	if (p) p[4]='#';
	p = strstr(tbmpname,"CLER\\");
	if (p) p[4]='#';
	DOSTranslate(tbmpname,bmpname);

	fhsiz = 14;//sizeof(BITMAPFILEHEADER);
	ihsiz = 40;//sizeof(BITMAPINFOHEADER);
	bmfh.bfType = 19778;
	bmfh.bfSize = fhsiz+ihsiz+256L*sizeof(RGBQUAD)+out->logw*out->logh;
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = fhsiz+ihsiz+256L*sizeof(RGBQUAD);

	bmih.biSize = ihsiz;
	bmih.biWidth = out->realw;
	bmih.biHeight = out->logh;
	bmih.biPlanes = 1;
	bmih.biBitCount = 8;
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = out->logw*out->logh;
	bmih.biXPelsPerMeter = 0;
	bmih.biYPelsPerMeter = 0;
	bmih.biClrUsed = 256;
	bmih.biClrImportant = 256;

 	st = fopen(bmpname,"wb");
	if (st!=NULL)
	{
		SafeWrite(&bmfh.bfType,sizeof(bmfh.bfType),1,st);
		SafeWrite(&bmfh.bfSize,sizeof(bmfh.bfSize),1,st);
		SafeWrite(&bmfh.bfReserved1,sizeof(bmfh.bfReserved1),1,st);
		SafeWrite(&bmfh.bfReserved2,sizeof(bmfh.bfReserved2),1,st);
		SafeWrite(&bmfh.bfOffBits,sizeof(bmfh.bfOffBits),1,st);

		SafeWrite(&bmih.biSize,sizeof(bmih.biSize),1,st);
		SafeWrite(&bmih.biWidth,sizeof(bmih.biWidth),1,st);
		SafeWrite(&bmih.biHeight,sizeof(bmih.biHeight),1,st);
		SafeWrite(&bmih.biPlanes,sizeof(bmih.biPlanes),1,st);
		SafeWrite(&bmih.biBitCount,sizeof(bmih.biBitCount),1,st);
		SafeWrite(&bmih.biCompression,sizeof(bmih.biCompression),1,st);
		SafeWrite(&bmih.biSizeImage,sizeof(bmih.biSizeImage),1,st);
		SafeWrite(&bmih.biXPelsPerMeter,sizeof(bmih.biXPelsPerMeter),1,st);
		SafeWrite(&bmih.biYPelsPerMeter,sizeof(bmih.biYPelsPerMeter),1,st);
		SafeWrite(&bmih.biClrUsed,sizeof(bmih.biClrUsed),1,st);
		SafeWrite(&bmih.biClrImportant,sizeof(bmih.biClrImportant),1,st);
		SafeWrite(pal,sizeof(RGBQUAD),256,st);

		for (i=0;i<out->logh;i++)
			SafeWrite(out->bmp+(out->logh-1-i)*out->logw,sizeof(BYTE),out->logw,st);

		fclose(st);
	}
	else dprintf(logfile,"Cannot open bitmap %s for write\n",bmpname);
}

// Reads BMP file with name translated from rbmpname
// Header is read for real width, and logical height
// Logical width is next multiple of 4 greater than or equal to real width
// Bitmap is written to short array *out, allocated to fit
// *out is allocated and read as logical width by logical height
//
// memory is allocated on *out
// max: logw*logh*sizeof(short)

void ReadBMP(Bitmap *out,char *rbmpname,RGBQUAD *pal)
{
	int i,n;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	FILE *st;
	char *buffer=NULL;
	char bmpname[256],tbmpname[256];
	char *p=NULL;

	strcpy(tbmpname,rbmpname);
	strupr(tbmpname);
	p = strstr(tbmpname,"VILE\\");
	if (p) p[4]='#';
	p = strstr(tbmpname,"CLER\\");
	if (p) p[4]='#';
	DOSTranslate(tbmpname,bmpname);

	out->logw = out->logh = out->realw = 0;
   	st = fopen(bmpname,"rb");
	if (st!=NULL)
	{
		fread(&bmfh.bfType,sizeof(bmfh.bfType),1,st);
		fread(&bmfh.bfSize,sizeof(bmfh.bfSize),1,st);
		fread(&bmfh.bfReserved1,sizeof(bmfh.bfReserved1),1,st);
		fread(&bmfh.bfReserved2,sizeof(bmfh.bfReserved2),1,st);
		fread(&bmfh.bfOffBits,sizeof(bmfh.bfOffBits),1,st);
		fread(&bmih.biSize,sizeof(bmih.biSize),1,st);
		fread(&bmih.biWidth,sizeof(bmih.biWidth),1,st);
		fread(&bmih.biHeight,sizeof(bmih.biHeight),1,st);
		fread(&bmih.biPlanes,sizeof(bmih.biPlanes),1,st);
		fread(&bmih.biBitCount,sizeof(bmih.biBitCount),1,st);
		fread(&bmih.biCompression,sizeof(bmih.biCompression),1,st);
		fread(&bmih.biSizeImage,sizeof(bmih.biSizeImage),1,st);
		fread(&bmih.biXPelsPerMeter,sizeof(bmih.biXPelsPerMeter),1,st);
		fread(&bmih.biYPelsPerMeter,sizeof(bmih.biYPelsPerMeter),1,st);
		fread(&bmih.biClrUsed,sizeof(bmih.biClrUsed),1,st);
		fread(&bmih.biClrImportant,sizeof(bmih.biClrImportant),1,st);
		fread(pal,sizeof(RGBQUAD),256,st);

		out->realw = bmih.biWidth;
		out->logw = 4*((bmih.biWidth+3)/4);
		out->logh = bmih.biHeight;

		n = out->logw*out->logh;

		buffer = malloc(n);
		fseek(st,bmfh.bfOffBits,SEEK_SET);
		fread(buffer,n,1,st);

		out->bmp = (BYTE *)malloc(n);

		for (i=out->logh-1;i>=0;i--)
			memmove
			(
				out->bmp+(out->logh-1-i)*out->logw,
				buffer+out->logw*i,
				out->logw
			);
		fclose(st);
		free(buffer);
	}
	else dprintf(logfile,"WARNING: bitmap for %s not supplied\n",tbmpname);
}

// Return the length of a picture structure as written in a wad
// no allocation

int GetPictureLen(PicData *p,int tallposts)
{
	int offs=0,i=0;
	unsigned char npix,rown;

	if (p->phgt>254 && !tallposts)
		return 4*sizeof(short)+p->pwid*sizeof(int)+p->pwid*(p->phgt+5);

	if (p->pwid>0)
	{
		offs = p->coloffs[p->pwid-1];
		i = offs - 4*sizeof(short) - p->pwid*sizeof(int);
		rown = p->posts[i++];
		while (rown!=255)
		{
			npix = p->posts[i++];
			i += npix+2;
			rown=p->posts[i++];
		}
	}
	return 4*sizeof(short)+p->pwid*sizeof(int)+i;										 
}

typedef struct stpost
{
	int rown;
	int coln;
	int pixcnt;
} Post;

static Post post1,post2;

int ReadPixel(Post *post, UBYTE **p, int wid)
{
	UBYTE *q = *p,ch;

restart0:
	if (post->pixcnt>0)
	{
		post->rown++;
		post->pixcnt--;
		ch = *q++;
		*p = q;
		if (ch==xparent) goto restart0;
		return ch;
	}
	else //if (post->pixcnt==0)
	{
		q++;
		while (post->coln<wid && *q==0xff)
		{
			post->coln++;
			q++;
		}
		if (post->coln>=wid) return -1;

		post->rown = *q++;
		post->pixcnt = *q++;
		q++;
		goto restart0;
	}
}

int CmpPicture(UBYTE *p1,UBYTE *p2)
{
	int w,h,c,d,o1,o2;

	if (memcmp(p1,p2,4*sizeof(short))) 	// if size/offset differ no compare
		return 2;

	w = ((short *)p1)[0];			  	// get size and offset
	h = ((short *)p1)[1];				// offset not used here

	o1 = ((long *)p1)[2];
	o2 = ((long *)p2)[2];

	post1.coln = post1.rown = 0; post1.pixcnt=0;
	post2.coln = post2.rown = 0; post2.pixcnt=0;
	p1 += o1-1;
	p2 += o2-1;

	while(1)
	{
		c=ReadPixel(&post1,&p1,w);
		d=ReadPixel(&post2,&p2,w);
		if (c!=d)
		{
			return 1;
		}
		else if (c==-1)
		{
			return 0;
		}
	}
}

static int ComparePicNumsFast(int b, int p)
{
	char *buff1=NULL,*buff2=NULL;
	int bn,pn,bo,po,r;

	if (b<0 && p<0) return 0;
	if (b<0) return -3;
	if (p<0) return -4;
	bn = bdir.dirp[b].reslen;
	bo = bdir.dirp[b].resptr;
	pn = pdir.dirp[p].reslen;
	po = pdir.dirp[p].resptr;

	if (bn==pn && pn==0) return 0;
	if (bn==0 || pn==0) return 2;

	buff1 = (char *)malloc(bn);
	fseek(bst,bo,SEEK_SET);
	fread(buff1,bn,1,bst);

	buff2 = (char *)malloc(pn);
	fseek(pst,po,SEEK_SET);
	fread(buff2,pn,1,pst);

	r = CmpPicture(buff1,buff2);
	free(buff1);
	free(buff2);
	return r;
}

static int ComparePicNumsGood(int b, int p)
{
	int bn,pn,bo,po,r;
	int x1,y1,x2,y2;
	Bitmap out1,out2;

	if (b<0 && p<0) return 0;
	if (b<0) return -3;
	if (p<0) return -4;
	bn = bdir.dirp[b].reslen;
	bo = bdir.dirp[b].resptr;
	pn = pdir.dirp[p].reslen;
	po = pdir.dirp[p].resptr;

	if (bn==pn && pn==0) return 0;
	if (bn==0 || pn==0) return 2;

	Wad2BMP(&out1,&x1,&y1,bn,bo,bst);
	Wad2BMP(&out2,&x2,&y2,pn,po,pst);

	if (out1.logw!=out2.logw || out1.logh!=out2.logh || x1!=x2 || y1!=y2)
	{
		free(out1.bmp);
		free(out2.bmp);
		return 2;
	}

	r = memcmp(out1.bmp,out2.bmp,out1.logw*out1.logh)? 1 : 0;
	free(out1.bmp);
	free(out2.bmp);
	return r;
}

int ComparePicNums(int b, int p)
{
	if (ComparePicsGood)
		return ComparePicNumsGood(b,p);
	else
		return ComparePicNumsFast(b,p);
}

// Convert a rectangular array of numbers to a DOOM format picture
// bytes 0-255 represent palette colors as usual, xparent is transparency.
//
// memory is allocated on p->coloffs and p->posts
// max: 256k+rwid actual: postlen+4*(1+rwid)

void ConvertToPicture(Bitmap *out,int x,int y,int tallposts,PicData *p)
{
	int i,j,n;
	int pixcnt;
	int postlen;

	p->pwid = out->realw;
	p->phgt = out->logh;
	p->pxoff = x;
  p->pyoff = y;

	p->coloffs = (long *)malloc(out->logw*sizeof(int));

	// Worse case is alternating transparencies with pixels down
	// one post per out->logh/2 pixels is 5*(out->logh+1)/2 bytes, 
	// plus one for the 0xff terminating the column.

	p->posts = (unsigned char *)malloc((1+5*(out->logh+1)/2)*out->logw);

	n=postlen=0;
	for (j=0;j<out->realw;j++)
	{
		p->coloffs[j] = 4*sizeof(short) + p->pwid*sizeof(int) + postlen;
		pixcnt = 0;
		for (i=0;i<out->logh;i++)
		{
			if (out->bmp[out->logw*i+j]==xparent)
			{
				if (pixcnt>0)
				{
					p->posts[postlen++]=0;		// add fill byte at end
					pixcnt=0;
				}
			}
			else
			{
				if (pixcnt==254 && tallposts) // new post at 254 unless not tallposts
				{
					p->posts[postlen++]=0;		// add fill byte at end
					pixcnt=0;
				}

				if (pixcnt==0) 							// start a post
				{
					p->posts[postlen++] = (out->logh>254 && !tallposts)? 0 : i;
					n = postlen++;
					p->posts[n] = (out->logh>254 && !tallposts)? 0 : 1;
					p->posts[postlen++] = 0;  // add fill byte at start
				}
				else if (out->logh<254 || tallposts)
					p->posts[n]++;

				// add pixel to current post

				p->posts[postlen++] = out->bmp[out->logw*i+j];
				pixcnt++;
			}
		}
		if (pixcnt>0)										// terminate last post
			p->posts[postlen++]=0;
		p->posts[postlen++]=255;				// terminate column
	}
	p->posts = (unsigned char *)realloc(p->posts,postlen);
}

void Wad2BMP(Bitmap *out,int *xof,int *yof,int n,int o,FILE *st)
{
	char *buffer=NULL;
	PicData pic;

	if (n>=0)
	{
		buffer = malloc(n);
		fseek(st,o,SEEK_SET);
		fread(buffer,n,1,st);

		out->realw = pic.pwid = ((short *)buffer)[0];
		out->logh = pic.phgt = ((short *)buffer)[1];
		out->logw = ((out->realw+3)>>2)<<2;
		*xof = pic.pxoff = ((short *)buffer)[2];
		*yof = pic.pyoff = ((short *)buffer)[3];
		pic.coloffs = (long *)(((short *)buffer)+4);
		pic.posts = (unsigned char *)(buffer+pic.coloffs[0]);

		out->bmp = (BYTE *) malloc(out->logw*out->logh);
		memset(out->bmp,xparent,out->logw*out->logh);

		OverlayPicture(out,&pic,0,0);
		free(buffer);
	}
}

// Write a picture structure as a lump in an opened file
// memory balanced

int WritePictureImage(FILE *st,int offs,PicData *p,int tallposts)
{
	int n;
	short *pchd;
	int *cola;
	unsigned char *post;
	char *img = NULL;

	img = (char *) malloc (n=GetPictureLen(p,tallposts));

	pchd = (short *)(img);
	cola = (int *)(((char *)pchd) + 4*sizeof(short));
	post = (unsigned char *)(((char *)cola) + p->pwid*sizeof(int));

	memcpy(pchd,&(p->pwid),4*sizeof(short));
	memcpy(cola,p->coloffs,(p->pwid)*sizeof(int));
	memcpy(post,p->posts,n-4*sizeof(short)-p->pwid*sizeof(int));

	fseek(st,offs,SEEK_SET);
	SafeWrite(img,n,1,st);
	free(img);

	return ftell(st);
}

// Read the 62 font character graphics resources from an IWAD
// no allocation

void InitDoomFont()
{
	int j,n,k;

	for (j=33;j<94;j++)
	{
		k = j-33;
		n = DoomFont[k].pwid = ((short *)FontData[k])[0];
		DoomFont[k].phgt = ((short *)FontData[k])[1];
		DoomFont[k].pxoff = ((short *)FontData[k])[2];
		DoomFont[k].pyoff = ((short *)FontData[k])[3];
		DoomFont[k].coloffs = (long *)(((short *)FontData[k])+4);
		DoomFont[k].posts = (unsigned char *)(DoomFont[k].coloffs+n);

		n = BigFont[k].pwid = ((short *)BigData[k])[0];
		BigFont[k].phgt = ((short *)BigData[k])[1];
		BigFont[k].pxoff = ((short *)BigData[k])[2];
		BigFont[k].pyoff = ((short *)BigData[k])[3];
		BigFont[k].coloffs = (long *)(((short *)BigData[k])+4);
		BigFont[k].posts = (unsigned char *)(BigFont[k].coloffs+n);
	}
}

// Closest match one palette with another, and create lookup table
// to translate palette indices of images

void MatchPalettes(RGBQUAD *curpal,RGBQUAD *outpal,int *lookup)
{
	int i,j,mn,d;

	for (i=0;i<256;i++)
	{
		if (i==xparent) continue;
		for (j=0,mn=200000000;j<256;j++)
		{
			if (j==xparent) continue;
			if ((d=colordist(curpal+i,outpal+j))<mn)
			{
				mn = d;
				lookup[i] = j;
			}
		}
	}
	lookup[xparent] = xparent;
}


static int mxcir[64];
static int mxhir[64];

// Break a text creation control string into segments delimited by changes in
// color, line number, or brightness
// no allocation

void SegmentText(Segment *segs,int *ns,char *text)
{

 	int i,e,nsegs=*ns;
	int isup=0,islw=0,isbg=0;
	char colstr[8]=" rbgwsy";
	char valnstr[8]="^-v";
	char halnstr[8]="<|>";
	char bristr[12]=" 01234567";
	PicData *fp;

 	for (i=0;i<64;i++)
 	{
 		mxcir[i]=0;
		mxhir[i]=0;
 	}

	nsegs=0;
 	segs[nsegs].color=TEXTRED;
 	segs[nsegs].bright=0;
 	segs[nsegs].fsel=0;
	segs[nsegs].valn=V_TOP;
	segs[nsegs].haln=H_CENTER;
 	segs[nsegs].row=0;
 	segs[nsegs].col=0;
 	segs[nsegs].ncols=0;
 	segs[nsegs].len=0;
 	segs[nsegs].str[0]='\0';

 	for (i=0;i<strlen(text);i++)
 	{
 		switch(text[i])
 		{
 			case '\\':
				switch(tolower(text[i+1]))
				{
					// color change escape code
					case 'r':
					case 'b':
					case 'g':
					case 's':
					case 'w':
					case 'y':
						e = strchr(colstr,text[i+1]) - colstr;
						if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
							mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
						nsegs++; 
						segs[nsegs].color = e-1;
						segs[nsegs].bright = segs[nsegs-1].bright;
						segs[nsegs].valn = segs[nsegs-1].valn;
						segs[nsegs].haln = segs[nsegs-1].haln;
						segs[nsegs].fsel = segs[nsegs-1].fsel;
						segs[nsegs].row = segs[nsegs-1].row;
						segs[nsegs].col = segs[nsegs-1].col+segs[nsegs-1].ncols;
						segs[nsegs].ncols = 0;
						segs[nsegs].len = 0;
						segs[nsegs].str[0]='\0';
						i++;
						break;

					// vertical alignment escapes
					case '^':
					case '-':
					case 'v':
						e = strchr(valnstr,text[i+1]) - valnstr;
						if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
							mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
						nsegs++; 
						segs[nsegs].color = segs[nsegs-1].color;
						segs[nsegs].bright = segs[nsegs-1].bright;
						segs[nsegs].valn = e;
						segs[nsegs].haln = segs[nsegs-1].haln;
						segs[nsegs].fsel = segs[nsegs-1].fsel;
						segs[nsegs].row = segs[nsegs-1].row;
						segs[nsegs].col = segs[nsegs-1].col+segs[nsegs-1].ncols;
						segs[nsegs].ncols = 0;
						segs[nsegs].len = 0;
						segs[nsegs].str[0]='\0';
						i++;
						break;

					// horizontal alignment escapes
					case '<':
					case '|':
					case '>':
						e = strchr(halnstr,text[i+1]) - halnstr;
						if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
							mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
						nsegs++; 
						segs[nsegs].color = segs[nsegs-1].color;
						segs[nsegs].bright = segs[nsegs-1].bright;
						segs[nsegs].haln = e;
						segs[nsegs].valn = segs[nsegs-1].valn;
						segs[nsegs].fsel = segs[nsegs-1].fsel;
						segs[nsegs].row = segs[nsegs-1].row;
						segs[nsegs].col = segs[nsegs-1].col+segs[nsegs-1].ncols;
						segs[nsegs].ncols = 0;
						segs[nsegs].len = 0;
						segs[nsegs].str[0]='\0';
						i++;
						break;

					// font select text escape code
					case 'f':
						e = isdigit(text[i+2])? text[i+++2]-'0' : 0;
						if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
							mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
						nsegs++; 
						segs[nsegs].color = segs[nsegs-1].color;
						segs[nsegs].bright = segs[nsegs-1].bright;
						segs[nsegs].valn = segs[nsegs-1].valn;
						segs[nsegs].haln = segs[nsegs-1].haln;
						segs[nsegs].fsel = e;
						segs[nsegs].row = segs[nsegs-1].row;
						segs[nsegs].col = segs[nsegs-1].col+segs[nsegs-1].ncols;
						segs[nsegs].ncols = 0;
						segs[nsegs].len = 0;
						segs[nsegs].str[0]='\0';
						i++;
						break;

					// brightness adjust escape code
					case '0':
					case '1':
					case '2':
					case '3':
					case '4':
					case '5':
					case '6':
					case '7':
						e = strchr(bristr,text[i+1]) - bristr;
						if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
							mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
						nsegs++; 
						segs[nsegs].color = segs[nsegs-1].color;
						segs[nsegs].bright = e-1;
						segs[nsegs].valn = segs[nsegs-1].valn;
						segs[nsegs].haln = segs[nsegs-1].haln;
						segs[nsegs].fsel = segs[nsegs-1].fsel;
						segs[nsegs].row = segs[nsegs-1].row;
						segs[nsegs].col = segs[nsegs-1].col+segs[nsegs-1].ncols;
						segs[nsegs].ncols = 0;
						segs[nsegs].len = 0;
						segs[nsegs].str[0]='\0';
						i++;
						break;

					// new line escape code
					case 'n':
						if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
							mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
						nsegs++; 
						segs[nsegs].color = segs[nsegs-1].color;
						segs[nsegs].bright = segs[nsegs-1].bright;
						segs[nsegs].valn = segs[nsegs-1].valn;
						segs[nsegs].haln = segs[nsegs-1].haln;
						segs[nsegs].fsel = segs[nsegs-1].fsel;
						segs[nsegs].row = segs[nsegs-1].row+1;
						segs[nsegs].col = 0;
						segs[nsegs].ncols = 0;
						segs[nsegs].len = 0;
						segs[nsegs].str[0]='\0';
						i++;
						break;

					// unknown escape code literalizes escape character
					default:
						// literalized character
						fp = segs[nsegs].fsel? BigFont : DoomFont;
						if (33<=text[i+1] && text[i+1]<=95)
						{
							segs[nsegs].ncols += fp[text[i+1]-33].pwid;
							if (fp[text[i+1]-33].phgt > mxhir[segs[nsegs].row])
								mxhir[segs[nsegs].row] = fp[text[i+1]-33].phgt;
							segs[nsegs].str[segs[nsegs].len++] = text[i+1];
							segs[nsegs].str[segs[nsegs].len] = '\0';
						}
						else if (text[i+1]!=13 && text[i+1]!=10)
						{
							segs[nsegs].ncols += 4;
							segs[nsegs].str[segs[nsegs].len++] = ' ';
							segs[nsegs].str[segs[nsegs].len] = '\0';
						}
						i++;
						break;
				}
				break;

			default:
				// regular text character
				fp = segs[nsegs].fsel? BigFont : DoomFont;

				if (FollowCase)
				{
					isup = isupper(text[i])? 1 : 0;
					islw = islower(text[i])? 1 : 0;
					isbg = (fp==BigFont)? 1 : 0;

					if ((isbg && islw) || (!isbg && isup))
					{
						if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
							mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
						nsegs++; 
						segs[nsegs].color = segs[nsegs-1].color;
						segs[nsegs].bright = segs[nsegs-1].bright;
						segs[nsegs].valn = segs[nsegs-1].valn;
						segs[nsegs].haln = segs[nsegs-1].haln;
						segs[nsegs].fsel = isup;
						fp = segs[nsegs].fsel? BigFont : DoomFont;
						segs[nsegs].row = segs[nsegs-1].row;
						segs[nsegs].col = segs[nsegs-1].col+segs[nsegs-1].ncols;
						segs[nsegs].ncols = 0;
						segs[nsegs].len = 0;
						segs[nsegs].str[0]='\0';
					}
				}

				text[i] = toupper(text[i]);
				if (33<=text[i] && text[i]<=95)
				{
					segs[nsegs].ncols += fp[text[i]-33].pwid;
					if (fp[text[i]-33].phgt > mxhir[segs[nsegs].row])
						mxhir[segs[nsegs].row] = fp[text[i]-33].phgt;
					segs[nsegs].str[segs[nsegs].len++] = text[i];
					segs[nsegs].str[segs[nsegs].len] = '\0';
				}
				else if (text[i]!=13 && text[i]!=10)
				{
					segs[nsegs].ncols += 4;
					segs[nsegs].str[segs[nsegs].len++] = ' ';
					segs[nsegs].str[segs[nsegs].len] = '\0';
				}
				break;
			
 		}
 	}
	if (segs[nsegs].col+segs[nsegs].ncols>mxcir[segs[nsegs].row])
		mxcir[segs[nsegs].row] = segs[nsegs].col+segs[nsegs].ncols;
	nsegs++;
	*ns = nsegs;
}

// create a PicData structure from a text creation control string
// memory is allocated on outpp->coloffs and outpp->posts by the
// call to ConvertPicture at the end

void RenderString(int targwid,char *strs,Bitmap *out)
{
	int i,j,k;
	int o,h,w,x,y;
	int coff=0,colpos;
	int rowhgt[1024];
	int maxh,maxc,toth;
	PicData *fp;
	int nrows,vsty;
	Segment *segs=NULL;
	int nsegs=0;
	Bitmap *b=out;

	segs = (Segment *)malloc(1024*sizeof(Segment));

	CenterText=1;
	DoubleSize=0;
	SegmentText(segs,&nsegs,strs);
	if (nsegs<=0) return;

	if (Verbose)
	{
		dprintf(logfile,"%s text segments:\n",strs);
		for (j=0;j<nsegs;j++)
			dprintf(logfile,"%d %d %d %d %d %d %s\n",j,segs[j].row,segs[j].col,segs[j].ncols,segs[j].color,segs[j].len,segs[j].str);
	}

	nrows = segs[nsegs-1].row+1;

	maxh=maxc=0;
	for (i=0,maxc=0;i<nrows;i++)
		if (mxcir[i]>maxc)
			maxc=mxcir[i];
	for (i=0,maxh=toth=0;i<nrows;i++)
	{
		if (mxhir[i]>maxh)
			maxh=mxhir[i];
		toth += mxhir[i]+4;
		rowhgt[i] = i>0? rowhgt[i-1]+mxhir[i-1]+4 : 0;
	}

	b->realw = b->logw = NextPowerTwo(max(maxc,targwid));
	b->logh = NextPowerTwo(toth);

	b->bmp = malloc(b->logw*b->logh);
	memset(b->bmp,xparent,b->logw*b->logh);

	vsty=H_CENTER;
	for (j=0;j<nsegs;j++)
	{
		for (k=j+1;k<nsegs && segs[k].row==segs[j].row;k++)
			vsty = segs[k].haln;
		switch(vsty)
		{
			case H_LEFT:
				coff = 0;
				break;
			case H_CENTER:
				coff = (b->logw-mxcir[segs[j].row])/2;
				break;
			case H_RIGHT:
				coff = (b->logw-mxcir[segs[j].row]);
				break;
		}
		colpos = segs[j].col+coff;
		XlateTextColor=segs[j].color;
		Brightness=segs[j].bright;
		fp = segs[j].fsel? BigFont : DoomFont;
		for (i=0;i<segs[j].len;i++)
		{
			if (33<=segs[j].str[i] && segs[j].str[i]<95)
     	{

	    	w = fp[segs[j].str[i]-33].pwid;
	    	h = fp[segs[j].str[i]-33].phgt;
        o = fp[segs[j].str[i]-33].pyoff;
				x = colpos;
				y = rowhgt[segs[j].row]-o;

				if (h<mxhir[segs[j].row])
				{
					if (segs[j].valn==V_BOTTOM)
						y += mxhir[segs[j].row]-h+o;
					else if (segs[j].valn==V_MIDDLE)
						y += (mxhir[segs[j].row]-h+o)/2;
				}

				Xlation=1;
	    	OverlayPicture(b,&fp[segs[j].str[i]-33],x,y);
				Xlation=0;
	    	colpos += w;
     	}
     	else colpos += 4;
		}
	}
	free(segs);
}

// Change the color and brightness of a text color
// no allocation

int XlateColor(int color,int xlate,int bright)
{
	int c;

	c=color-176;
	if (c<0 || c>15) c=15;

	switch(xlate)
	{
		case TEXTRED:
			c = redidx[c];
			break;
		case TEXTBLUE:
			c = blueidx[c];
			break;
		case TEXTGREEN:
			c = greenidx[c];
			break;
		case TEXTWHITE:
			c = whiteidx[c];
			break;
		case TEXTGRAY:
			c = grayidx[c];
			break;
		case TEXTYELLOW:
			c = yellowidx[c];
			break;
	}
	return ColorMap[4*bright][c];
}

void DoubleBitMap(BYTE **out,int *logw,int *logh)
{
	int i,j;
	int w,h;
	BYTE *dp;

	w = (*logw)<<1;
	h = (*logh)<<1;
	dp = *out = (BYTE *) realloc(*out, w*h);
	for (i=h-1;i>=0;i--)
		for (j=w-1;j>=0;j--)
			dp[i*w+j] = dp[(i/2)*(w/2)+j/2];
	*logw = w;
	*logh = h;
}

