#include "patch.h"

static int thingchk[4096];
static int linechk[16384];

/* Check Map m for Thing t intersecting Linedef l */

// killough 3/15/98:
//
// A very fast function for testing intersections between things and linedefs.

int ThingIsOnLine(Thing *t, Linedef *l)
{
  int dx = Vertexes[l->to].x - Vertexes[l->from].x;
  int dy = Vertexes[l->to].y - Vertexes[l->from].y;
  int a = Vertexes[l->from].x - t->x;    // Thing-->from vector
  int b = Vertexes[l->from].y - t->y;
  int r;
	r = Radius(*t);                          // Thing radius

  // First make sure bounding boxes of linedef and thing intersect.
  // Leads to quick rejection using only shifts and adds/subs/compares.
  if (abs(a*2+dx)-abs(dx) > 2*r || abs(b*2+dy)-abs(dy) > 2*r)
    return 0;

  //if (((r < a && r < a+dx) || (-r > a && -r > a+dx)) ||
  //  ((r < b && r < b+dy) || (-r > b && -r > b+dy)))
  //return 0;

  // Next, make sure that at least one thing crosshair intersects linedef.
  // Requires only 3-4 multiplications, the rest adds/subs/shifts/xors.
  // (writing the steps out this way leads to better code generation)

  a *= dy;
  b *= dx;
  a -= b;
  b = dx + dy;
  b *= r;
  if (((a-b)^(a+b)) < 0)
    return 1;
  dy -= dx;
  dy *= r;
  b = a+dy; 
  a -= dy;
  return (a^b) < 0;
}
// killough 3/15/98: end ThingIsOnLine()

int aThingIsOnLine(Thing *t,Linedef *l)
{
	int x1,y1,x2,y2;
	int r,xt,yt;

	double h,k,rr;

	/*
	Now suppose a line segment is described by vertices V1=(x1,y1) and
	V2=(x2,y2). Suppose a thing is at T=(xt,yt) and has radius r.
	*/
	x1 = Vertexes[l->from].x;
	y1 = Vertexes[l->from].y;
	x2 = Vertexes[l->to].x;
	y2 = Vertexes[l->to].y;
	xt = t->x;
	yt = t->y;
	r = Radius(*t);
	rr = r;	/* double precision copy */

	/*
	The four line segments about the thing are described by the
	vertex pairs:
		(xt-r,yt+r) (xt+r,yt+r)
		(xt-r,yt-r) (xt+r,yt-r)
		(xt-r,yt-r) (xt-r,yt+r)
		(xt+r,yt-r) (xt+r,yt+r) 
	*/

	/*
	First if (x1,y1) and (x2,y2) both lie within the bounding box
	intersection is guarenteed.
	*/
	if (xt-r <= x1 && x1 <= xt+r && 
	    yt-r <= y1 && y1 <= yt+r &&
	    xt-r <= x2 && x2 <= xt+r && 	
	    yt-r <= y2 && y2 <= yt+r) return 1;
 
	/*
	Suppose y1=y2. Then the linedef is horizontal. In order for an
	intersection to occur it is necessary and sufficient that y1 lies
	between yt+r and yt-r, and that the interval x1,x2 has a non-empty
	intersection with (xt-r, xt+r).  
	*/
	if (y1==y2)
	{
	    if (y1<yt-r || y1>yt+r) return 0;
	    if (x1<xt-r && x2<xt-r) return 0;
	    if (x1>xt+r && x2>xt+r) return 0;
	    return 1;
	}

	/*
	Similarly suppose x1=x2. Then the linedef is vertical. In order
	for an intersection to occur it is necessary and sufficient that
	x1 lies	between xt+r and xt-r, and that the interval y1,y2 has a
	non-empty intersection with yt+r,yt-r.
	*/
	if (x1==x2)
	{
	    if (x1<xt-r || x1>xt+r) return 0;
	    if (y1<yt-r && y2<yt-r) return 0;
	    if (y1>yt+r && y2>yt+r) return 0;
	    return 1;
	}

	/*
	Below assume y1<>y2 and x1<>x2
	Let P1 and P2 be points on the thing boundary & linedef
	P1 = (xt-r,yt+r) + h*(2r,0)  P2 = (x1,y1) + k*(x2-x1,y2-y1)
	We want 0<=h,k<=1 s.t. P1=P2 for an intersection to occur
	We test 
	*/

	k = (yt+rr-y1)/(y2-y1);
	if (0<=k && k<=1)
	{
		h = (k*(x2-x1) - (xt-rr-x1))/2*rr;
		if (0<=h && h<=1) return 1;
	}
	k = (yt-rr-y1)/(y2-y1);
	if (0<=k && k<=1)
	{
		h = (k*(x2-x1) - (xt-rr-x1))/(2*rr);
		if (0<=h && h<=1) return 1;
	}
	k = (xt-rr-x1)/(x2-x1);
	if (0<=k && k<=1)
	{
		h = (k*(y2-y1) - (yt-rr-y1))/(2*rr);
		if (0<=h && h<=1) return 1;
	}
	k = (xt+rr-x1)/(x2-x1);
	if (0<=k && k<=1)
	{
		h = (k*(y2-y1) - (yt-rr-y1))/(2*rr);
		if (0<=h && h<=1) return 1;
	}
	return 0;
}

/* Check for lines thru a Thing t on Map m */

int CheckLinesForThing(int tno)
{
	int i,s1,s2,sref1,sref2,f1,f2,c1,c2;
	int bad=0;

	for (i=0;i<NLinedefs;i++)
	{
		if (ThingIsOnLine(&Things[tno],&Linedefs[i]))
		{
			s1 = Linedefs[i].sides[0];
			s2 = Linedefs[i].sides[1];
			if (s1>=0 && s2>=0 /*Linedefs[i].attr&TwoSFlag*/)
			{
				sref1 = Sidedefs[s1].sect;
				sref2 = Sidedefs[s2].sect;
				f1 = Sectors[sref1].fhgt;
				f2 = Sectors[sref2].fhgt;
				c1 = Sectors[sref1].chgt;
				c2 = Sectors[sref2].chgt;
				if (
					EdgeTest &&
					abs(f1-f2)>24 &&
					!(ThingProps(Things[tno].typ)&FLY) &&
					!(ThingProps(Things[tno].typ)&HANG)
				   )
				{
					dprintf(logfile,"%s %d is balanced on linedef %d\n",Name(Things[tno].typ),tno,i);
					bad |= 4;
					linechk[i]|=4;
				}
				if (ThingProps(Things[tno].typ)&OBST)
				{
					if ((c1-f1 < Height(Things[tno])) && CeilingTest)
					{
						dprintf(logfile,"%s %d is partly squished under sector ceiling %d\n",Name(Things[tno].typ),tno,sref1);
						bad |= 8;
						linechk[i]|=8;
					}
					if ((c2-f2 < Height(Things[tno])) && CeilingTest)
					{
						dprintf(logfile,"%s %d is partly squished under sector ceiling %d\n",Name(Things[tno].typ),tno,sref2);
						bad |= 8;
						linechk[i]|=8;
					}
				}
			}
			else if (WallTest)
			{
				dprintf(logfile,"%s %d is stuck in wall (linedef) %d\n",Name(Things[tno].typ),tno,i);
				bad |= 2;
				linechk[i]|=2;
			}
		}
	}
	return bad;
}

/* Check Map m for improperly placed things */

void CheckMap(char *mapname,WadDir *dir,FILE *st)
{
	int i,j;
	int r,dx,dy,s;
	int bad=0;

	if (ReadLevelData(mapname,dir,st,NLevObj()))
	{
		FreeLevelData();
		return;
	}


  /* Check for poorly placed things  */

	/*
		Stuck together:
	 	Two things appearing on a common difficulty level, both with
	 	obstruction type, s.t. the sum of their radii is greater than
	 	both the x and y distances between them.

		Stuck in wall:
		Parts of a thing lie on both sides of a 1S linedef
	    segment in the wad.

		Balanced on edge:
		Parts of a thing that doesn't hang or fly lie on both sides
		of a 2S linedef which spans a floor height difference greater
		than 24.

		Squished:
		Parts of a thing lie in a sector whose floor to ceiling height
		difference is inadequate to support the thing. This condition
		can prevent floors there or elsewhere from being lowered if the
		thing is an obstruction.
	*/

	for (i=0;i<NLinedefs;i++)
		linechk[i]=0;

	for (i=0;i<NThings;i++)
	{
        thingchk[i]=0;
		if (!MultiPlayer && Things[i].attr&mplyrmask) continue;
		if (MonstersOnly && !(ThingProps(Things[i].typ)&MONST)) continue;
		if (CeilingTest)
		{
			s = SectorOf(Things[i].x,Things[i].y);
			if (s>=0 && Sectors[s].chgt-Sectors[s].fhgt < Height(Things[i]))
			{
				if (ThingProps(Things[i].typ)&OBST)
				{
					dprintf(logfile,"%s %d is squished under sector ceiling %d\n",Name(Things[i].typ),i,s);
					bad |= 8;
                    thingchk[i] |= 8;
				}
			}
		}
		if (TogetherTest)
		{
			for (j=0;j<i;j++)
			{
				if (
						(Things[i].attr&Things[j].attr&7)
						&&
						(ThingProps(Things[i].typ)&OBST)
						&&
						(ThingProps(Things[j].typ)&OBST)
			   	)
				{
					dx = abs(Things[i].x-Things[j].x);
					dy = abs(Things[i].y-Things[j].y);
					r = Radius(Things[i])+Radius(Things[j]);
					if (dx<r &&	dy<r)
					{
						dprintf(logfile,"%s %d and %s %d are stuck together\n",Name(Things[i].typ),i,Name(Things[j].typ),j);
						bad |= 1;
                        thingchk[i] |= 1;
					}
				}
			}

		}
		if (EdgeTest || CeilingTest || WallTest)
        {
            j = CheckLinesForThing(i);
			bad |= j;
            thingchk[i] |= j;
        }
	}
	if (MonstersOnly)
	{
		if (CeilingTest && !(bad&8)) dprintf(logfile,"No monsters squished by ceiling\n");
		if (EdgeTest && !(bad&4)) dprintf(logfile,"No monsters balanced on edges\n");
		if (WallTest && !(bad&2)) dprintf(logfile,"No monsters stuck in walls\n");
		if (TogetherTest && !(bad&1)) dprintf(logfile,"No monsters stuck together\n");
	}
	else
	{
		if (CeilingTest && !(bad&8)) dprintf(logfile,"No things squished by ceilings\n");
		if (EdgeTest && !(bad&4)) dprintf(logfile,"No things balanced on edges\n");
		if (WallTest && !(bad&2)) dprintf(logfile,"No things stuck in walls\n");
		if (TogetherTest && !(bad&1)) dprintf(logfile,"No things stuck together\n");
	}
	FreeLevelData();
}

