#include "patch.h"

//////////////////////////////////////////////
//------------------------------------------//
// == Angle routines =======================//
//------------------------------------------//
//////////////////////////////////////////////

//////////////////////////////////////////////////////////////
//
// PointToAngle - logical routine to translate
// an ordered pair of coordinates to a binary angle
//
// return the 16 bit unsigned angle the line 
// from the first to the second point makes with the x-axis.
//
unsh PointToAngle(short x1,short y1,short x2,short y2)
{
	register short dy = y2-y1;										// dy/dx slope of line
	register short dx = x2-x1;
	double ang = atan2(dy,dx);										// angle between -pi and pi
																								// of line with x-axis (east)
	if (ang<0) ang+=2*pi;
	if (ang>=2*pi) ang-=2*pi;											// adjust for 0 to 2*pi range
	return  (unsh)(65536*(ang/(2*pi)));           // convert to 0 to 65535
}

///////////////////////////////////////////////////////////////////////////
//
// VertNum2Angle - doom routine to translate an ordered pair
// of vertex numbers to a 16 bit unsigned angle 
//
// Convert a direction specified by an ordered pair of vertex numbers
// to a 16 bit unsigned DOOM angle, as used by Segs
//
unsh VertNum2Angle(unsh from,unsh to)
{
	return
		PointToAngle
		(
			Vertexes[from].x,
			Vertexes[from].y,
			Vertexes[to].x,
			Vertexes[to].y
		);
}

//////////////////////////////////////////////
//------------------------------------------//
// == Side of line a position has ==========//
//------------------------------------------//
//////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
//
// PointLineSideValue - logical routine 
//
// given a line from P(lx1,ly1) to Q(lx2,ly2) and a point V(x,y) return
// how far to the left of the line Q-P V is. (Q-P) X (V-P) is the distance
// to V from the line Q-P times the length of (Q-P). Note that the V s.t.
// (Q-P)X(V-P)=0, are the points along the line. When this quantity is
// negative we are to the left of the line, when positive to its right.
// Left and right being defined as you stand at P and look towards Q.
//
// In terms of coordinates Q-P X V-P = (y-ly1)(lx2-lx1) - (x-lx1)(ly2-ly1)
// Divide this by length of (Q-P) to obtain distance from point to line.
// 

int PointLineSideValue(short x,short y,short lx1,short ly1,short lx2,short ly2)
{
	// Q-P = (lx2-lx1,ly2-ly1)    V-P = (x-lx1,y-ly1)
	return //lineside(x,y,lx1,ly1,lx2,ly2);
	 (y-ly1) * (int)(lx2-lx1) - (ly2-ly1) * (int)(x-lx1); // (Q-P) X (V-P)
}

///////////////////////////////////////////////////////////////////////////
//
// PointLineSideCompare
//
// return the logical relation of a point and a line. Return SD_LEFT (1)
// if on the left, SD_ON (0) if on the line and SD_RIGHT (-1) if the
// right of the line.
//

int PointLineSideCompare(short x,short y,short lx1,short ly1,short lx2,short ly2)
{
	int s= PointLineSideValue(x,y,lx1,ly1,lx2,ly2);
	return s>0? SD_LEFT : (s<0? SD_RIGHT : SD_ON);  
}

////////////////////////////////////////////////////////////////////////
//
// PointOnLineSide
//
// Compare the position (x,y) with the line from
// (lx1,ly1) to (lx2,ly2). Return 1 if on the
// line or its right side, return 0 if on the
// left side of the line.
//

int PointOnLineSide(short x,short y,short lx1,short ly1,short lx2,short ly2)
{
  return PointLineSideValue(x,y,lx1,ly1,lx2,ly2) >= 0;
}

////////////////////////////////////////////////////////////////////////
//
// PointOnNodeSide
//
// Given a position and a BSP node return whether the position is on the
// right or left side of the node division line.
//

int PointOnNodeSide(short x, short y, Node *node)
{
  return (y-node->y)*(int)node->dx - (x-node->x)*(int)node->dy >= 0;
}

// Given a position and a BSP seg return whether the position is on the
// right (1) or left (0) side of the seg

int PointOnSegSide(short x, short y, Seg *sg)
{
  return
		PointOnLineSide(x,y,X1Of(sg),Y1Of(sg),X2Of(sg),Y2Of(sg));
}

int PointSegSideCompare(short x,short y,Seg *sg)
{
  return
		PointLineSideCompare(x,y,X1Of(sg),Y1Of(sg),X2Of(sg),Y2Of(sg));
}

int PointSegSideValue(short x,short y,Seg *sg)
{
	return
		PointLineSideValue(x,y,X1Of(sg),Y1Of(sg),X2Of(sg),Y2Of(sg));
}

//////////////////////////////////////////////
//------------------------------------------//
// == Line routines ========================//
//------------------------------------------//
//////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// LineCompare - compare a line segment ST with a line PQ. Return the side
// of the line each vertex of the segment lies. Also return whether
// the segment and line are in opposite directions or not.
//
// If the vertexes of the segment line on opposite sides of the line
// an intersection exists. In that case return the distance along the
// the segment from each vertex to the intersection.
//
// The return value is true if an intersection exists
//
// Line: P(x1,y1),Q(x2,y2) Segment: S(w1,z1),T(w2,z2) 
//
//	dx1 = x2 - x1, dy1 = y2 - y1;				// Q-P = (dx1,dy1)
//	dx2 = w2 - w1, dy2 = z2 - z1;				// T-S = (dx2,dy2)
//	dx3 = w2 - x1, dy3 = z2 - y1;				// T-P = (dx3,dy3)
//	dx4 = dx3-dx2, dy4 = dy3-dy2;				// S-P = (dx4,dy4)

SegCmpTypes LineCompare
(
	int dx1,int dy1,							  // Q-P
	int dx2,int dy2,								// T-S
	int dx3,int dy3,								// T-P
	long *fromrel,									// signed dist S to PQ
	long *torel											// signed dist T to PQ
)
{
	long d = dx1*dy2 - dx2*dy1;			// T-S X Q-P
	long t = dy3*dx1 - dy1*dx3;			// T-P X Q-P   >0 means left of
	long f = t-d;										// S-P X Q-P	 >0 means left of

	// in the case the lines intersect, change f,t to true distances

	if ((f^t)<0 && f && t)
	{
		long dist = (((int)sqrt(dx2*dx2+dy2*dy2))<<16)/(labs(f)+labs(t));

		f = (f*dist)>>16;							// make these actual signed distances
		t = (t*dist)>>16;

		if (labs(f)<=2) f=0;					// call anything within 2 a hit
		if (labs(t)<=2) t=0;
	}

	*fromrel = f;										// now set the return values
	*torel = t;

	// return a symbol classifying interaction of line and segment

	if ((f^t)<0 && f && t)					// seg has ends on opposite sides of line
		return SEG_SPLIT;
	else if (!f && !t)							// seg and line are conincident
	{
		long dir = dx1*dx2 + dy1*dy2;	// (Q-P)*(T-S) >0 means same dir
		return dir>=0? SEG_RCOINC : SEG_LCOINC;
	}
	else if (f<0 || t<0)						// seg is on right of line
		return SEG_RIGHT;
	else 														// seg is on left of line
		return SEG_LEFT;
}

////////////////////////////////////////////////////////////////////////
//
// LinedefCompare - compare a line segment with a line. Return the side
// of the line each vertex of the segment lies. Also return whether
// the segment and line are in opposite directions or not.
//
// If the vertexes of the segment line on opposite sides of the line
// an intersection exists. In that case return the distance along the
// the segment from each vertex to the intersection.
//
// The return value is true if an intersection exists
//

int LinedefCompare(unsh nfrom1,unsh nto1,unsh nfrom2,unsh nto2,long *fromrel,long *torel)
{
	Vertex *fr1=Vertexes+nfrom1;
	Vertex *fr2=Vertexes+nfrom2;
	Vertex *to1=Vertexes+nto1;
	Vertex *to2=Vertexes+nto2;

	int dx1=to1->x-fr1->x, dy1=to1->y-fr1->y;
	int dx2=to2->x-fr2->x, dy2=to2->y-fr2->y;
	int dx3=to2->x-fr1->x, dy3=to2->y-fr1->y;

	return LineCompare(dx1,dy1,dx2,dy2,dx3,dy3,fromrel,torel);
}

Ssector *PointInSubsector(short x, short y)
{
	Ssector *s;
  unsigned short nodenum = NNodes-1;
	int i;

  while (!(nodenum & NF_SUBSECTOR))
    nodenum = Nodes[nodenum].children[PointOnNodeSide(x,y,Nodes+nodenum)];

  s = Ssectors + (nodenum & ~NF_SUBSECTOR);
	for (i=0;i<s->nsegs;i++)
	{
		Vertex *v1=V1Of(SscSeg(s)+i); //Vertexes+Segs[s->seg0+i].from;
		Vertex *v2=V2Of(SscSeg(s)+i); //Vertexes+Segs[s->seg0+i].to;
		if (PointLineSideValue(x,y,v1->x,v1->y,v2->x,v2->y)>=16)
			return NULL;
	}
	return s;
}

Ssector *PointNearSubsector(short x, short y)
{
  unsigned short nodenum = NNodes-1;

  while (!(nodenum & NF_SUBSECTOR))
    nodenum = Nodes[nodenum].children[PointOnNodeSide(x,y,Nodes+nodenum)];

  return Ssectors + (nodenum & ~NF_SUBSECTOR);
}

Sector *PointInSector(short x,short y)
{
	return SegSec(SscSeg(PointInSubsector(x,y)));
}

Sector *PointNearSector(short x,short y)
{
	return SegSec(SscSeg(PointNearSubsector(x,y)));
}

Sector *SubSectorInSector(Ssector *ssec)
{
	return SscSec(ssec);
}

/* Return euclidean distance */

double Dist(double x1,double y1,double x2,double y2)
{
	return hypot(x1-x2,y1-y2);
}

double VertsToLen(Vertex *v1, Vertex *v2)
{
	return Dist(v1->x,v1->y,v2->x,v2->y);
}

/* Return length of linedef */

double LineLen(Linedef *l)
{
  return hypot(DXOf(l),DYOf(l));
}

/* Return length of segment */

double SegLen(Seg *s)
{
  return hypot(DXOf(s),DYOf(s));
}


/* Return distance from point to Linedef */

double Ldist(double x, double y, Linedef l)
{
	double px,py,dx,dy,rx,ry,d,den,s0;

 	px = Vertexes[l.from].x;
	py = Vertexes[l.from].y;
	dx = Vertexes[l.to].x - px;
	dy = Vertexes[l.to].y - py;

	den = dx*dx + dy*dy;
	s0 = (x-px)*dx + (y-py)*dy;
	s0 /= den;

	d = min(Dist(x,y,px,py),Dist(x,y,px+dx,py+dy));
	if (s0>0 && s0<1) 
	{
		rx = s0*dx + px;
		ry = s0*dy + py;
	 	d = min(d,Dist(x,y,rx,ry));
	}
	return d;
}

int Lnear(Linedef l,Vertex v)
{
	int x1,y1,x2,y2;
	int x,y;

	x1 = Vertexes[l.from].x;
	y1 = Vertexes[l.from].y;
	x2 = Vertexes[l.to].x;
	y2 = Vertexes[l.to].y;
	x = v.x; y = v.y;

	return (y-y1)*(x2-x1) - (x-x1)*(y2-y1) > 16? 0 : 1;
}

/* Find the nearest Linedef to a point */

int NearestLinedef(int x,int y)
{
 	double dx,dy,ld,mind;
	int i,mini=-1;

	dx = x; dy = y;
	mind = 100000;
	for (i=0;i<NLinedefs;i++)
	{
		ld = Ldist(dx,dy,Linedefs[i]);
		if (ld<mind)
		{
			mini = i;
			mind = ld;
		}
	}
	return mini;
}


int NearestBlkLinedef(int x,int y)
{
 	short org_x,org_y;
	int nrows,ncols;
	int r,mini,i,j,k,col,row;
	double ld,mind;

	if (!BlockmapValid())
		return NearestLinedef(x,y);

	org_x = ((short *)Blockmap)[0];
	org_y = ((short *)Blockmap)[1];
	ncols = Blockmap[2];
	nrows = Blockmap[3];

	col = (x-org_x)>>7;
	row = (y-org_y)>>7;
	mini = -1;
	mind = 100000;
	for (r = 0; r < max(nrows,ncols); r++)
	{
		int inc = r? 2*r : 1;
		int r0 = max(0,row-r);
		int r1 = min(nrows-1,row+r);
		int c0 = max(0,col-r);
		int c1 = min(ncols-1,col+r);
		for (j=r0;j<=r1;j+=inc)
		{
			for (i=c0;i<=c1;i+=inc)
			{
				unsigned short *bp = BlockmapPtr(i,j);
				if (bp)
				{
					for (k=0;bp[k]!=0xffff;k++)
					{
						ld = Ldist(x,y,Linedefs[bp[k]]);
						if (ld<mind)
						{
							mini = bp[k];
							mind = ld;
						}
					}
				}
			}
		}
		if (mini>=0) break;
	}
	return mini;
}

int SubSectorOf(int x,int y)
{
	return PointInSubsector(x,y)-Ssectors;
}

/* Find and report the sector a point lies in */

int SectorOf(int x, int y)
{
	int n;
	double fx,fy,tx,ty,s;
	double v1x,v1y,v2x,v2y;
	Linedef l;
	Sidedef sd;


	/* get the number of the nearest linedef to the point */

	n = NearestBlkLinedef(x,y);
	if (n>=0) l = Linedefs[n];
	else return -1;

	/* determine which sidedef of the linedef is facing the point */
 	fx = Vertexes[l.from].x;
	fy = Vertexes[l.from].y;
	tx = Vertexes[l.to].x;
	ty = Vertexes[l.to].y;
	v1x = x - fx;
	v1y = y - fy;
	v2x = tx - fx;
	v2y = ty - fy;
	s = v2y*v1x - v1y*v2x; /* cross product with unit-z vector */

	/* return the sector reference of the sideddef */
	if (s >= 0 && l.sides[0]>=0) sd = Sidedefs[l.sides[0]];
	else if (s <= 0 && l.sides[1]>=0) sd = Sidedefs[l.sides[1]];
	else return -1;

	return sd.sect;
}

/* Find and report the nearest line to a point, within a threshold */

int LineOf(int x, int y,double thresh)
{
	int n;

	/* get the number of the nearest linedef to the point */
	n = NearestBlkLinedef(x,y);

	if (n>=0)
		if (Ldist(x,y,Linedefs[n])<=thresh)
			return n;
	return -1;
}

void RenumberLine(int ln)
{
	Linedef l;
	int i;

	if (ln<=0) return; // already done or not a line!
	if (ln>=NLinedefs) return; // not a line!

	// Make a copy of ln-th line
	l = Linedefs[ln];

	// Move lines below ln-th up one
	for (i=ln;i>0;i--)
		Linedefs[i] = Linedefs[i-1];

	// Make renumbered line we copied linedef #0
	Linedefs[0] = l;
}

void ClearSecrets()
{
	int i;

	for (i=0;i<NSectors;i++)
		if (Sectors[i].typ==9) Sectors[i].typ=0;
}

void GetVertexesInSubSector(int subsect,int **objs,int *nobjs)
{
	Ssector *sbs = Ssectors+subsect;
	int ns = sbs->nsegs;
	int s0 = sbs->seg0;
	int i;

	for (i=s0;i<s0+ns;i++)
	{
		Seg *sg = Segs+i;

		AddDynInt(objs,nobjs,sg->from);
		AddDynInt(objs,nobjs,sg->to);
	}
}

void GetVertexesInSector(int sect,int **objs,int *nobjs)
{
	int j;

	for (j=0;j<NLinedefs;j++)
	{
		Linedef *ld = Linedefs+j;
		int s1 = ld->sides[0];
		int s2 = ld->sides[1];

		if
		(
			(s1>=0 && Sidedefs[s1].sect==sect)
			||
			(s2>=0 && Sidedefs[s2].sect==sect)
		)
		{
			AddDynInt(objs,nobjs,ld->from);
			AddDynInt(objs,nobjs,ld->to);
		}
	}
}

void GetVertexesInTaggedSectors(int tag,int **objs,int *nobjs)
{
	int i,j;

	if (tag<=0)
		return;

	for (i=0;i<NSectors;i++)
		if (Sectors[i].tag==tag)
			for (j=0;j<NLinedefs;j++)
			{
				Linedef *ld = Linedefs+j;
				int s1 = ld->sides[0];
				int s2 = ld->sides[1];

				if
				(
					(s1>=0 && Sidedefs[s1].sect==i)
					||
					(s2>=0 && Sidedefs[s2].sect==i)
				)
				{
					AddDynInt(objs,nobjs,ld->from);
					AddDynInt(objs,nobjs,ld->to);
				}
			}
}

void GetVertexesWithTag(int tag,int **objs,int *nobjs)
{
	int i;

	if (tag<=0)
		return;
	for (i=0;i<NLinedefs;i++)
	{
		Linedef *ld = Linedefs+i;
		if (ld->tag==tag)
		{
			AddDynInt(objs,nobjs,ld->from);
			AddDynInt(objs,nobjs,ld->to);
		}
	}
}

void GetLinesInSector(int sect,int **objs,int *nobjs)
{
	int j,s1,s2;

	for (j=0;j<NLinedefs;j++)
		if
		(
			(
				(s1=Linedefs[j].sides[0])>=0
				&&
				Sidedefs[s1].sect==sect
			)
			||
			(
				(s2=Linedefs[j].sides[1])>=0
				&&
				Sidedefs[s2].sect==sect
			)
		)
			AddDynInt(objs,nobjs,j);
}

void GetLinesInTaggedSectors(int tag,int **objs,int *nobjs)
{
	int i,j,s1,s2;

	if (tag<=0) return;
	for (i=0;i<NSectors;i++)
		if (Sectors[i].tag==tag)
			for (j=0;j<NLinedefs;j++)
				if
				(
					(
						(s1=Linedefs[j].sides[0])>=0
						&&
						Sidedefs[s1].sect==i
					)
					||
					(
						(s2=Linedefs[j].sides[1])>=0
						&&
						Sidedefs[s2].sect==i
					)
				)
					AddDynInt(objs,nobjs,j);
}

void GetLinesWithTag(int tag,int **objs,int *nobjs)
{
	int i;

	if (tag<=0) return;
	for (i=0;i<NLinedefs;i++)
		if (Linedefs[i].tag==tag)
			AddDynInt(objs,nobjs,i);
}

int LinedefTouchesBox(int ln,int minx,int miny,int maxx,int maxy)
{
	int v1x,v1y,v2x,v2y,r;
	double dbvy,dbvx,dvy,dvx,dbx,dby,det,s,t;

	// the four line segments comprising the box
	// L0: (minx,miny) + t*(maxx-minx,0)
	// L1: (minx,miny) + t*(0,maxy-miny)
	// L2: (maxx,maxy) - t*(maxx-minx,0)
	// L3: (maxx,maxy) - t*(0,maxy-miny)
	//
	// The intersection with a generalized line segment
	// (v1x,v1y) to (v2x,v2y)
	// (v1x,v1y) + t*(v2x-v1x,v2y-v1y)
	//
	// Intersection of line segment with L0
	// --------------------------------------------------------------------
	//  v1x+t*(v2x-v1x) = minx + s*(maxx-minx)   v1y+t*(v2y-v1y) = miny
	//
	//	v1x+(miny-v1y)*(v2x-v1x)/(v2y-v1y) = minx + s*(maxx-minx);
	//  (miny-v1y)*(v2x-v1x) = (minx-v1x)*(v2y-v1y) + s*(maxx-minx)*(v2y-v1y)
	//
	// if v2y==v1y then int only if v2y==miny. check v1x,v2x in minx-maxx
	// else
	//                              ((miny-v1y)*(v2x-v1x)-(minx-v1x)*(v2y-v1y))
	// t= (miny-v1y)/(v2y-v1y)  s=  -------------------------------------------
	//                                      (maxx-minx)*(v2y-v1y)
	// int iff 0<=t<=1 and 0<=s<=1
	// 

	v1x = Vertexes[Linedefs[ln].from].x;
	v2x = Vertexes[Linedefs[ln].to].x;
	v1y = Vertexes[Linedefs[ln].from].y;
	v2y = Vertexes[Linedefs[ln].to].y;
	dvy = v2y-v1y;
	dvx = v1x-v2x;
	dbx = maxx-minx;
	dby = maxy-miny;

	// L0 intersects Linedef[ln]
	dbvy = miny-v1y;
	dbvx = minx-v1x;
	det = dbvy*dvx-dbvx*dvy;

	if (!dvy)
	{
		if (v1y!=miny) r = 0;
		else r = (v1x>=minx && v1x<=maxx) || (v1x>=minx && v1x<=maxx);
	}
	else
	{
		t = dbvy/dvy; s = det/(dbx*dvy);
		r = 0<=t && t<=1 && 0<=s && s<=1;
	}
	if (r) return r;

	// L2 intersects Linedef[ln]
	dbvy = maxy-v1y;
	dbvx = minx-v1x;
	det = dbvy*dvx-dbvx*dvy;

	if (!dvy)
	{
		if (v1y!=maxy) r=0;
		else r = ((v1x>=minx && v1x<=maxx) || (v1x>=minx && v1x<=maxx));
	}
	else
	{
		t = dbvy/dvy; s = det/(dbx*dvy);
		r = 0<=t && t<=1 && 0<=s && s<=1;
	}
	if (r) return 1;

	// Intersection of line segment with L1
	// --------------------------------------------------------------------
	//  v1y+t*(v2y-v1y) = miny + s*(maxy-miny)   v1x+t*(v2x-v1x) = minx
	//
	//	v1y+(minx-v1x)*(v2y-v1y)/(v2x-v1x) = miny + s*(maxy-miny);
	//  (minx-v1x)*(v2y-v1y) = (miny-v1y)*(v2x-v1x) + s*(maxy-miny)*(v2x-v1x)
	//
	// if v2x==v1x then int only if v2x==minx. check v1y,v2y in miny-maxy
	// else
	//                              ((minx-v1x)*(v2y-v1y)-(miny-v1y)*(v2x-v1x))
	// t= (minx-v1x)/(v2x-v1x)  s=  -------------------------------------------
	//                                      (maxy-miny)*(v2x-v1x)
	// int iff 0<=t<=1 and 0<=s<=1
	// 

	// L1 intersects Linedef[ln]
	dbvy = miny-v1y;
	dbvx = minx-v1x;
	det = dbvx*dvy-dbvy*dvx;

	if (!dvx)
	{
		if (v1x!=minx) r = 0;
		else r = (v1y>=miny && v1y<=maxy) || (v1y>=miny && v1y<=maxy);
	}
	else
	{
		t = dbvx/dvx; s = det/(dby*dvx);
		r = 0<=t && t<=1 && 0<=s && s<=1;
	}
	if (r) return r;

	// L3 intersects Linedef[ln]
	dbvy = miny-v1y;
	dbvx = maxx-v1x;
	det = dbvx*dvy-dbvy*dvx;

	if (!dvx)
	{
		if (v1x!=minx) r = 0;
		else r = (v1y>=miny && v1y<=maxy) || (v1y>=miny && v1y<=maxy);
	}
	else
	{
		t = dbvx/dvx; s = det/(dby*dvx);
		r = 0<=t && t<=1 && 0<=s && s<=1;
	}

	if (r) return r;
	return 0;
}
